from .convert import (
    AsNumpyArray,
    ExpandDimArray,
    VstackFeatures,
    ConcatFeatures,
    SwapAxes,
    ListFeatures,
    TargetDimIndicator,
    SampleTargetDim,
    CDFtoGaussianTransform,
    cdf_to_gaussian_forward_transform,
)
from .dataset import TransformedDataset
from .feature import (
    target_transformation_length,
    AddObservedValuesIndicator,
    AddConstFeature,
    AddTimeFeatures,
    AddAgeFeature,
)
from .field import (
    RemoveFields,
    RenameFields,
    SetField,
    SetFieldIfNotPresent,
    SelectFields,
)
from .sampler import (
    InstanceSampler,
    UniformSplitSampler,
    TestSplitSampler,
    ExpectedNumInstanceSampler,
    BucketInstanceSampler,
    ContinuousTimePointSampler,
    ContinuousTimeUniformSampler,
)
from .split import (
    shift_timestamp,
    InstanceSplitter,
    CanonicalInstanceSplitter,
    ContinuousTimeInstanceSplitter,
)
from .transform import (
    Transformation,
    Chain,
    Identity,
    MapTransformation,
    SimpleTransformation,
    AdhocTransform,
    FlatMapTransformation,
    FilterTransformation,
)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import functools
import glob
import random
from pathlib import Path
from typing import Iterator, List
from typing import NamedTuple

import rapidjson as json

from .common import Dataset, DataEntry, SourceContext
from .process import ProcessDataEntry


def load(file_obj):
    for line in file_obj:
        yield json.loads(line)


class Span(NamedTuple):
    path: Path
    line: int


class Line(NamedTuple):
    content: object
    span: Span


class JsonLinesFile:
    """
    An iterable type that draws from a JSON Lines file.

    Parameters
    ----------
    path
        Path of the file to load data from. This should be a valid
        JSON Lines file.
    """

    def __init__(self, path: Path, shuffle: bool = True) -> None:
        self.path = path
        self.shuffle = shuffle

    def __iter__(self):
        with open(self.path) as jsonl_file:
            lines = jsonl_file.read().splitlines()
            if self.shuffle:
                random.shuffle(lines)

            for line_number, raw in enumerate(lines, start=1):
                span = Span(path=self.path, line=line_number)
                try:
                    yield Line(json.loads(raw), span=span)
                except ValueError:
                    raise Exception(
                        f"Could not read json line {line_number}, {raw}"
                    )

    def __len__(self):
        # 1MB
        BUF_SIZE = 1024**2

        with open(self.path) as file_obj:
            read_chunk = functools.partial(file_obj.read, BUF_SIZE)
            return sum(chunk.count("\n") for chunk in iter(read_chunk, ""))


class FileDataset(Dataset):
    """
    Dataset that loads JSON Lines files contained in a path.

    Parameters
    ----------
    path
        Return list of path names that match path. Each file is considered
        and should be valid. A valid line in a file can be for
        instance: {"start": "2014-09-07", "target": [0.1, 0.2]}.
    freq
        Frequency of the observation in the time series.
        Must be a valid Pandas frequency.
    one_dim_target
        Whether to accept only univariate target time series.
    shuffle
        Whether to shuffle the time series when making the batches
    """

    def __init__(
        self,
        path: Path,
        freq: str,
        one_dim_target: bool = True,
        shuffle: bool = False,
    ) -> None:
        self.shuffle = shuffle
        self.path = path
        self.process = ProcessDataEntry(freq, one_dim_target=one_dim_target)
        if not self.files():
            raise OSError(f"no valid file found via {path}")

    def __iter__(self) -> Iterator[DataEntry]:
        for path in self.files():
            for line in JsonLinesFile(path, self.shuffle):
                data = self.process(line.content)
                data["source"] = SourceContext(
                    source=line.span.path, row=line.span.line
                )
                yield data

    def __len__(self):
        return sum([len(JsonLinesFile(path)) for path in self.files()])

    def files(self) -> List[Path]:
        """
        List the files that compose the dataset.

        Returns
        -------
        List[Path]
            List of the paths of all files composing the dataset.
        """
        files = glob.glob(str(self.path))
        if self.shuffle:
            random.shuffle(files)
        return files



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import torch.nn
import torch.optim

from .distr_tcn import DistributionalTCN


def train_step_from_batch(
    ts_chunks: torch.Tensor,
    targets: torch.Tensor,
    distr_tcn: DistributionalTCN,
    optimizer: torch.optim.Adam,
):
    """
    Arguments
    ----------
    ts_chunks: Mini-batch chunked from the time series
    targets: Corresponding chunk of target values
    distr_tcn: DistributionalTCN
    otimizer: Optimizer containing parameters, learning rate, etc
    """
    distr_outputs = distr_tcn(ts_chunks.float())
    loss = -distr_outputs.log_prob(targets.float())
    loss = loss.mean()
    loss_value = loss.cpu().detach().numpy()
    loss.backward()
    optimizer.step()
    return loss_value


def eval_on_series(
    distr_tcn: DistributionalTCN,
    optimizer: torch.optim.Adam,
    series_tensor: torch.Tensor,
    ts_len: int,
    context_length: int,
    is_train: bool = False,
    return_predictions: bool = False,
    lead_time: int = 1,
):
    """
    Arguments
    ----------
    distr_tcn: DistributionalTCN
    otimizer: Optimizer containing parameters, learning rate, etc
    series_tensor: Time series
    ts_len: Length of time series
    context_length: Number of time steps to input
    is_train: True if time series is training set
    return_predictions: True if to return (loss, predictions), False if to return loss only
    lead_time: Number of time steps to predict ahead
    """
    loss_log = []

    # Parallelising the training:
    trying_mini_batches = True
    if is_train:
        mini_batch_size = 64
        stride = 1

        window_length = context_length + lead_time

        unfold_layer = torch.nn.Unfold(
            kernel_size=(1, window_length), stride=stride
        )

        ts_windows = (
            unfold_layer(series_tensor.unsqueeze(2))
            .transpose(1, 2)
            .transpose(0, 1)
        )

        numb_mini_batches = ts_windows.shape[0] // mini_batch_size

        if trying_mini_batches:
            batch_indices = np.arange(ts_len - window_length - 1)
            for i in range(numb_mini_batches):
                idx = np.random.choice(batch_indices, mini_batch_size)
                batch_indices = np.setdiff1d(batch_indices, idx)

                ts_chunks = ts_windows[idx, :, :-lead_time]
                targets = ts_windows[idx, :, -1]

                loss_log.append(
                    train_step_from_batch(
                        ts_chunks, targets, distr_tcn, optimizer
                    )
                )

        else:
            ts_chunks = ts_windows[:, :, :-lead_time]
            targets = ts_windows[:, :, -1]

            loss_log.append(train_step_from_batch(ts_chunks, targets))

        return loss_log

    if return_predictions:
        predictions = {
            "low_lower": [],
            "lower": [],
            "median": [],
            "upper": [],
            "up_upper": [],
        }

    for i in range(ts_len - context_length - lead_time - 1):
        ts_chunk = series_tensor[:, :, i : i + context_length]
        target = series_tensor[:, :, i + context_length + lead_time - 1]

        distr_output = distr_tcn(ts_chunk.float())

        if return_predictions:
            predictions["lower"].append(distr_output.icdf(torch.tensor(0.05)))
            predictions["median"].append(distr_output.icdf(torch.tensor(0.5)))
            predictions["upper"].append(distr_output.icdf(torch.tensor(0.95)))
            predictions["low_lower"].append(
                distr_output.icdf(torch.tensor(0.01))
            )
            predictions["up_upper"].append(
                distr_output.icdf(torch.tensor(0.99))
            )

        loss = -distr_output.log_prob(target.float())
        loss_value = loss.cpu().detach().numpy()[0]
        loss_log.append(loss_value)

        if is_train:
            loss.backward()
            optimizer.step()

    if return_predictions:
        return loss_log, predictions
    return loss_log


def plot_prediction(
    val_ts_tensor: torch.Tensor,
    predictions: torch.Tensor,
    context_length: int,
    lead_time: int = 1,
    start: int = 0,
    end: int = 500,
    fig: Optional[matplotlib.figure.Figure] = None,
):
    """
    Arguments
    ----------
    val_ts_tensor: Time series
    predictions: Prediction series
    context_length: Number of time steps to input
    lead_time: Number of time steps to predict ahead
    start: Index of time series at which to start plotting
    end: Index of time series at which to end plotting
    """
    end = np.minimum(end, len(predictions["lower"]))

    for key in predictions.keys():
        try:
            predictions[key] = np.array(
                list(map(lambda x: x.detach().cpu().numpy(), predictions[key]))
            ).ravel()
        except Exception:
            # TODO: handle error
            pass

    if fig is None:
        fig = plt.figure(figsize=(16, 7))
    plt.plot(
        val_ts_tensor.cpu().flatten()[context_length + lead_time - 1 :][
            start:end
        ],
        "ko",
        markersize=2,
        label=r"$x_t$",
    )
    plt.plot(
        predictions["up_upper"][start:end],
        "r-",
        alpha=0.3,
        label=r"$z_{0.01}(t)|x_{t-1}$",
    )
    plt.plot(
        predictions["upper"][start:end],
        "y-",
        alpha=0.3,
        label=r"$z_{0.05}(t)|x_{t-1} \rightarrow \tau_t^{\rm{upper}}$",
    )
    plt.fill_between(
        np.arange(start, end),
        predictions["lower"][start:end],
        predictions["upper"][start:end],
        facecolor="y",
        alpha=0.3,
        label=None,
    )
    plt.plot(
        predictions["median"][start:end], "g", label=r"$z_{0.5\,}(t)|x_{t-1}$"
    )
    plt.plot(
        predictions["lower"][start:end],
        "y-",
        alpha=0.3,
        label=r"$z_{0.95}(t)|x_{t-1} \rightarrow \tau_t^{\rm{lower}}$",
    )
    plt.plot(
        predictions["low_lower"][start:end],
        "r-",
        alpha=0.3,
        label=r"$z_{0.99}(t)|x_{t-1}$",
    )

    plt.legend(loc="upper right")

    median = np.array(predictions["median"][start:end])
    true_val = np.array(
        val_ts_tensor.cpu().flatten()[context_length + lead_time - 1 :][
            start:end
        ]
    )
    MAE = np.mean(np.abs(median - true_val))
    plt.title(str(MAE))

    return fig


def highlight_min(data, color="lightgreen"):
    """
    Highlights the minimum in a Series or DataFrame
    """
    attr = "background-color: {}".format(color)
    if data.ndim == 1:  # Series from .apply(axis=0) or axis=1
        is_min = data == data.min()
        return [attr if v else "" for v in is_min]
    else:  # from .apply(axis=None)
        return pd.DataFrame(
            np.where(is_min, attr, ""), index=data.index, columns=data.columns
        )



# -*- coding: utf-8 -*-
import time
from contextlib import contextmanager
from io import StringIO

import numpy as np
import torch


class Timer:
    """
    Timer for PyTorch code
    Comes in the form of a contextmanager:

    Example:
    >>> timer = Timer()
    ... for i in range(10):
    ...     with timer("expensive operation"):
    ...         x = torch.randn(100)
    ... print(timer.summary())
    """

    def __init__(self, verbosity_level=1, skip_first=True, on_cuda=True):
        self.verbosity_level = verbosity_level
        # self.log_fn = log_fn if log_fn is not None else self._default_log_fn
        self.skip_first = skip_first
        self.cuda_available = torch.cuda.is_available() and on_cuda

        self.reset()

    def reset(self):
        """Reset the timer"""
        self.totals = {}  # Total time per label
        self.first_time = {}  # First occurrence of a label (start time)
        self.last_time = {}  # Last occurence of a label (end time)
        self.call_counts = {}  # Number of times a label occurred

    @contextmanager
    def __call__(self, label, epoch=-1.0, verbosity=1):
        # Don't measure this if the verbosity level is too high
        if verbosity > self.verbosity_level:
            yield
            return

        # Measure the time
        self._cuda_sync()
        start = time.time()
        yield
        self._cuda_sync()
        end = time.time()

        # Update first and last occurrence of this label
        if label not in self.first_time:
            self.first_time[label] = start
        self.last_time[label] = end

        # Update the totals and call counts
        if label not in self.totals and self.skip_first:
            self.totals[label] = 0.0
            del self.first_time[label]
            self.call_counts[label] = 0
        elif label not in self.totals and not self.skip_first:
            self.totals[label] = end - start
            self.call_counts[label] = 1
        else:
            self.totals[label] += end - start
            self.call_counts[label] += 1

    def _cuda_sync(self):
        """Finish all asynchronous GPU computations to get correct timings"""
        if self.cuda_available:
            torch.cuda.synchronize()

    def _default_log_fn(self, _, values, tags):
        label = tags["label"]
        epoch = values["epoch"]
        duration = values["value"]
        print(f"Timer: {label:30s} @ {epoch:4.1f} - {duration:8.5f}s")



import pandas as pd
import pytest

from pts.dataset import ProcessStartField


@pytest.mark.parametrize(
    "freq, expected",
    [
        ("B", "2019-11-01"),
        ("W", "2019-11-03"),
        ("M", "2019-11-30"),
        ("12M", "2019-11-30"),
        ("A-DEC", "2019-12-31"),
    ],
)
def test_process_start_field(freq, expected):
    process = ProcessStartField.process
    given = "2019-11-01 12:34:56"

    assert process(given, freq) == pd.Timestamp(expected, freq)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=missing-function-docstring
import logging
import pickle
import tempfile
from pathlib import Path
from typing import Optional
import numpy as np
import torch
from sacred import Experiment
from tsbench.analysis import EnsembleAnalyzer
from tsbench.config import MODEL_REGISTRY
from tsbench.constants import DEFAULT_DATA_PATH, DEFAULT_EVALUATIONS_PATH
from tsbench.evaluations.tracking import ModelTracker
from tsbench.forecasts import EnsembleWeighting

ex = Experiment()


@ex.config
def experiment_config():
    # pylint: disable=unused-variable
    experiment = "test"  # type: ignore
    data_path = str(DEFAULT_DATA_PATH)  # type: ignore
    evaluations_path = str(DEFAULT_EVALUATIONS_PATH)  # type: ignore

    weighting = "uniform"  # type: ignore
    size = 10  # type: ignore
    model_class = None  # type: ignore


@ex.automain
def main(
    _seed: int,
    data_path: str,
    evaluations_path: str,
    weighting: EnsembleWeighting,
    size: int,
    model_class: Optional[str],
):
    np.random.seed(_seed)
    torch.manual_seed(_seed)
    logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)

    # First, get the tracker
    print("Fetching the data...")
    tracker = ModelTracker.from_directory(
        Path(evaluations_path), data_path=Path(data_path)
    )

    # Evaluate the ensemble that can be built
    print("Evaluating the ensemble...")
    evaluator = EnsembleAnalyzer(
        tracker,
        ensemble_size=size,
        ensemble_weighting=weighting,
        config_class=MODEL_REGISTRY[model_class]
        if model_class is not None
        else None,
    )
    df, configs = evaluator.run()

    # Eventually, we store the results
    print("Storing the results...")
    with tempfile.TemporaryDirectory() as d:
        df_path = Path(d) / "results.parquet"
        df.to_parquet(str(df_path))
        ex.add_artifact(df_path, content_type="application/octet-stream")

        config_path = Path(d) / "configs.pickle"
        with config_path.open("wb+") as f:
            pickle.dump(configs, f)
        ex.add_artifact(config_path, content_type="application/octet-stream")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from itertools import islice
from pathlib import Path

import pytorch_lightning as pl

from gluonts.dataset.common import ListDataset
from gluonts.dataset.repository.datasets import get_dataset
from gluonts.model.predictor import Predictor
from gluonts.torch.model.deepar import DeepAREstimator


def test_torch_deepar():
    constant = get_dataset("constant")

    estimator = DeepAREstimator(
        freq=constant.metadata.freq,
        prediction_length=constant.metadata.prediction_length,
        batch_size=4,
        num_batches_per_epoch=3,
        trainer_kwargs=dict(max_epochs=2),
    )

    predictor = estimator.train(
        training_data=constant.train,
        validation_data=constant.train,
        shuffle_buffer_length=5,
    )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    forecasts = predictor_copy.predict(constant.test)

    for f in islice(forecasts, 5):
        f.mean


def test_torch_deepar_with_features():
    freq = "1h"
    prediction_length = 12

    training_dataset = ListDataset(
        [
            {
                "start": "2021-01-01 00:00:00",
                "target": [1.0] * 200,
                "feat_static_cat": [0, 1],
                "feat_static_real": [42.0],
                "feat_dynamic_real": [[1.0] * 200] * 3,
            },
            {
                "start": "2021-02-01 00:00:00",
                "target": [1.0] * 100,
                "feat_static_cat": [1, 0],
                "feat_static_real": [1.0],
                "feat_dynamic_real": [[1.0] * 100] * 3,
            },
        ],
        freq=freq,
    )

    prediction_dataset = ListDataset(
        [
            {
                "start": "2021-01-01 00:00:00",
                "target": [1.0] * 200,
                "feat_static_cat": [0, 1],
                "feat_static_real": [42.0],
                "feat_dynamic_real": [[1.0] * (200 + prediction_length)] * 3,
            },
            {
                "start": "2021-02-01 00:00:00",
                "target": [1.0] * 100,
                "feat_static_cat": [1, 0],
                "feat_static_real": [1.0],
                "feat_dynamic_real": [[1.0] * (100 + prediction_length)] * 3,
            },
        ],
        freq=freq,
    )

    estimator = DeepAREstimator(
        freq=freq,
        prediction_length=prediction_length,
        batch_size=4,
        num_batches_per_epoch=3,
        num_feat_dynamic_real=3,
        num_feat_static_real=1,
        num_feat_static_cat=2,
        cardinality=[2, 2],
        trainer_kwargs=dict(max_epochs=2),
    )

    predictor = estimator.train(
        training_data=training_dataset,
        validation_data=training_dataset,
        shuffle_buffer_length=5,
    )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    forecasts = predictor_copy.predict(prediction_dataset)

    for f in islice(forecasts, 5):
        f.mean



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, cast, Dict, List, Optional, Set, Union
import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.feature_extraction import DictVectorizer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline, make_union
from sklearn.preprocessing import (
    MinMaxScaler,
    PowerTransformer,
    StandardScaler,
)
from tsbench.config import (
    Config,
    EnsembleConfig,
    MODEL_REGISTRY,
    ModelConfig,
    TrainConfig,
)
from tsbench.config.model.models import SeasonalNaiveModelConfig
from tsbench.evaluations.tracking import ModelTracker


class ConfigTransformer(TransformerMixin):
    """
    The config transformer transforms a configuration (model + dataset) into a real-valued vector.
    """

    def __init__(
        self,
        add_model_features: bool = True,
        add_dataset_statistics: bool = True,
        add_seasonal_naive_performance: bool = False,
        add_catch22_features: bool = False,
        tracker: Optional[ModelTracker] = None,
    ):
        """
        Args:
            add_model_type: Whether a one-hot encoding of the model type as well as model
                hyperparameters should be added.
            add_dataset_statistics: Whether simple dataset statistics ought to be added.
            add_seasonal_naive_performance: Whether to add the nCRPS performance of Seasonal Naïve.
                Requires the cacher to be set.
            add_catch_22_features: Whether a dataset's catch22 features ought to be added.
            tracker: An optional tracker to obtain the performance of Seasonal Naïve.
        """
        assert any(
            [
                add_model_features,
                add_dataset_statistics,
                add_seasonal_naive_performance,
                add_catch22_features,
            ]
        ), "ConfigTransformer must be given at least some group of features."
        assert (
            not add_seasonal_naive_performance or tracker is not None
        ), "Tracker must be set if seasonal naive performance is used."

        self.encoders = []
        if add_model_features:
            self.encoders.append(ModelEncoder(transform_full_config=True))
        if add_dataset_statistics:
            self.encoders.append(DatasetStatisticsEncoder())
        if add_seasonal_naive_performance and tracker is not None:
            self.encoders.append(SeasonalNaivePerformanceEncoder(tracker))
        if add_catch22_features:
            self.encoders.append(DatasetCatch22Encoder())

        self.pipeline = make_union(*self.encoders)

    @property
    def feature_names_(self) -> List[str]:
        """
        Returns the feature names for the columns of transformed configurations.
        """
        return [
            f
            for e in self.encoders
            for f in e.feature_names_
            if isinstance(e, Encoder)
        ]

    def fit(self, X: List[Config[ModelConfig]]) -> ConfigTransformer:
        """
        Uses the provided configurations to fit the transformer pipeline.

        Args:
            X: The input configurations.
        """
        self.pipeline.fit(X)
        return self

    def transform(
        self, X: List[Config[ModelConfig]]
    ) -> npt.NDArray[np.float32]:
        """
        Transforms the given configurations according to the fitted transformer pipeline.

        Args:
            X: The input configurations.

        Returns:
            A NumPy array of shape [N, D]. N is the number of input configurations, D the dimension
                of the vectorized representation.
        """
        return cast(npt.NDArray[np.float32], self.pipeline.transform(X))


class EnsembleConfigTransformer(TransformerMixin):
    """
    The config transformer transforms a set of model configs into a set of NumPy arrays.
    """

    model_type_map: Dict[str, int]
    attribute_map: Dict[str, int]

    def __init__(self):
        self.scaler = StandardScaler()
        self.model_type_map = {}
        self.attribute_map = {}

    @property
    def feature_names_(self) -> List[str]:
        """
        Returns the feature names for the columns of transformed configurations.
        """
        return [f"model={m}" for m in self.model_type_map] + list(
            self.attribute_map.keys()
        )

    def fit(
        self, X: List[Config[EnsembleConfig]]
    ) -> EnsembleConfigTransformer:
        """
        Uses the provided configurations to fit the transformer pipeline.

        Args:
            X: The input configurations.
        """
        configs = [model for ensemble in X for model in ensemble.model]

        # Extract model types and attributes
        model_types = {c.name() for c in configs}
        self.model_type_map = {k: i for i, k in enumerate(sorted(model_types))}

        # pylint: disable=no-member
        shared_attributes = set(TrainConfig.__dataclass_fields__.keys())
        attributes = {
            k if k in shared_attributes else f"{c.name()}_{k}"
            for c in configs
            for k in c.__dataclass_fields__.keys()
        }
        self.attribute_map = {
            k: i + len(self.model_type_map)
            for i, k in enumerate(sorted(attributes))
        }

        # Transform the configs to fit the scaler for the attributes
        out = self._transform(configs)
        rhs = out[:, len(MODEL_REGISTRY) :]
        self.scaler.fit(rhs)
        return self

    def transform(
        self, X: List[Config[EnsembleConfig]]
    ) -> List[npt.NDArray[np.float32]]:
        """
        Transforms the given configurations according to the fitted transformer pipeline.

        Args:
            X: The input configurations.

        Returns:
            A NumPy array of shape [N, D] for every ensemble configuration where N is the number of
                ensemble members and D the dimensionality. N might differ for list members.
        """
        configs = [model for ensemble in X for model in ensemble.model]
        out = self._transform(configs)
        lhs = out[:, : len(MODEL_REGISTRY)]
        rhs = out[:, len(MODEL_REGISTRY) :]
        out = np.concatenate([lhs, self.scaler.transform(rhs)], axis=1)
        transformed = np.nan_to_num(out, nan=0)
        chunks = np.cumsum([0] + [len(ensemble.model) for ensemble in X])
        return [
            transformed[lower:upper]
            for lower, upper in zip(chunks, chunks[1:])
        ]

    def _transform(
        self, configs: List[ModelConfig]
    ) -> npt.NDArray[np.float32]:
        result = np.empty((len(configs), len(self.feature_names_)))
        result.fill(np.nan)

        # pylint: disable=no-member
        shared_attributes = set(TrainConfig.__dataclass_fields__.keys())
        for i, config in enumerate(configs):
            result[i, self.model_type_map[config.name()]] = 1
            for field in config.__dataclass_fields__:
                if field in shared_attributes:
                    result[i, self.attribute_map[field]] = getattr(
                        config, field
                    )
                else:
                    result[
                        i, self.attribute_map[f"{config.name()}_{field}"]
                    ] = getattr(config, field)

        return result


# -------------------------------------------------------------------------------------------------
# pylint: disable=missing-class-docstring,missing-function-docstring


class Encoder(ABC):
    @property
    @abstractmethod
    def feature_names_(self) -> List[str]:
        pass


class NanImputer:
    def fit(self, _X: List[Dict[str, Any]], _y: Any = None) -> NanImputer:
        return self

    def transform(
        self, X: List[Dict[str, Any]], _y: Any = None
    ) -> List[Dict[str, Any]]:
        df = pd.DataFrame(X)
        return df.to_dict("records")


class Selector:
    def __init__(
        self, use: Optional[Set[str]] = None, ignore: Optional[Set[str]] = None
    ):
        assert bool(use) != bool(
            ignore
        ), "One of `use` or `ignore` must be set."

        self.use = use or set()
        self.ignore = ignore or set()

    def fit(self, _X: List[Dict[str, Any]], _y: Any = None) -> Selector:
        return self

    def transform(
        self, X: List[Dict[str, Any]], _y: Any = None
    ) -> List[Dict[str, Any]]:
        return [
            {
                k: v
                for k, v in item.items()
                if (not self.use or k in self.use)
                and (not self.ignore or k not in self.ignore)
            }
            for item in X
        ]


# -------------------------------------------------------------------------------------------------


class ModelEncoder(Encoder):
    def __init__(self, transform_full_config: bool = False):
        self.transform_full_config = transform_full_config
        self.model_vectorizer = DictVectorizer(
            dtype=np.float32, sparse=False, sort=True
        )
        self.hp_vectorizer = DictVectorizer(
            dtype=np.float32, sparse=False, sort=True
        )
        self.pipeline = make_pipeline(
            NanImputer(),
            make_union(
                make_pipeline(
                    Selector(use={"model"}),
                    self.model_vectorizer,
                ),
                make_pipeline(
                    Selector(ignore={"model"}),
                    self.hp_vectorizer,
                    StandardScaler(),
                    SimpleImputer(strategy="constant", fill_value=0.0),
                ),
            ),
        )

    @property
    def feature_names_(self) -> List[str]:
        return (
            self.model_vectorizer.feature_names_
            + self.hp_vectorizer.feature_names_
        )

    def fit(
        self,
        X: Union[List[Config[ModelConfig]], List[ModelConfig]],
        _y: Any = None,
    ) -> ModelEncoder:
        if self.transform_full_config:
            self.pipeline.fit(
                [x.model.asdict() for x in cast(List[Config[ModelConfig]], X)]
            )
        else:
            self.pipeline.fit([x.asdict() for x in cast(List[ModelConfig], X)])
        return self

    def transform(
        self,
        X: List[Union[List[Config[ModelConfig]], List[ModelConfig]]],
        _y: Any = None,
    ) -> npt.NDArray[np.float32]:
        if self.transform_full_config:
            return cast(
                npt.NDArray[np.float32],
                self.pipeline.transform(
                    [
                        x.model.asdict()
                        for x in cast(List[Config[ModelConfig]], X)
                    ]
                ),
            )
        return cast(
            npt.NDArray[np.float32],
            self.pipeline.transform(
                [x.asdict() for x in cast(List[ModelConfig], X)]
            ),
        )


class DatasetStatisticsEncoder(Encoder):
    def __init__(self):
        self.unscaled_vectorizer = DictVectorizer(
            dtype=np.float32, sparse=False, sort=True
        )
        self.scaled_vectorizer = DictVectorizer(
            dtype=np.float32, sparse=False, sort=True
        )
        self.pipeline = make_union(
            make_pipeline(
                Selector(use={"integer_dataset", "frequency"}),
                self.unscaled_vectorizer,
            ),
            make_pipeline(
                Selector(ignore={"integer_dataset", "frequency"}),
                self.scaled_vectorizer,
                MinMaxScaler(),  # required for numerical stability
                PowerTransformer(),
            ),
        )

    @property
    def feature_names_(self) -> List[str]:
        return (
            self.unscaled_vectorizer.feature_names_
            + self.scaled_vectorizer.feature_names_
        )

    def fit(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> DatasetStatisticsEncoder:
        self.pipeline.fit(
            [
                {
                    **x.dataset.stats(),
                    "frequency": x.dataset.meta.freq,
                    "prediction_length": x.dataset.meta.prediction_length,
                }
                for x in X
            ]
        )
        return self

    def transform(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> npt.NDArray[np.float32]:
        return cast(
            npt.NDArray[np.float32],
            self.pipeline.transform([x.dataset.stats() for x in X]),
        )


class SeasonalNaivePerformanceEncoder(Encoder):
    def __init__(self, tracker: ModelTracker):
        self.tracker = tracker
        self.scaler = StandardScaler()

    @property
    def feature_names_(self) -> List[str]:
        return ["seasonal_naive_ncrps"]

    def fit(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> SeasonalNaivePerformanceEncoder:
        self.scaler.fit(self._get_performance_array(X))
        return self

    def transform(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> npt.NDArray[np.float32]:
        return cast(
            npt.NDArray[np.float32],
            self.scaler.transform(self._get_performance_array(X)),
        )

    def _get_performance_array(
        self, X: List[Config[ModelConfig]]
    ) -> npt.NDArray[np.float32]:
        return np.array(
            [
                self.tracker.get_performance(
                    Config(SeasonalNaiveModelConfig(), x.dataset)
                ).ncrps.mean
                for x in X
            ]
        )[:, None]


class DatasetCatch22Encoder(Encoder):
    def __init__(self):
        self.vectorizer = DictVectorizer(
            dtype=np.float32, sparse=False, sort=True
        )
        self.pipeline = make_pipeline(
            self.vectorizer,
            PowerTransformer(),
        )

    @property
    def feature_names_(self) -> List[str]:
        return self.vectorizer.feature_names_

    def fit(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> DatasetCatch22Encoder:
        datasets = {x.dataset for x in X}
        features = {d: d.catch22().mean().to_dict() for d in datasets}
        self.pipeline.fit([features[x.dataset] for x in X])
        return self

    def transform(
        self, X: List[Config[ModelConfig]], _y: Any = None
    ) -> npt.NDArray[np.float32]:
        datasets = {x.dataset for x in X}
        features = {d: d.catch22().mean().to_dict() for d in datasets}
        return cast(
            npt.NDArray[np.float32],
            self.pipeline.transform([features[x.dataset] for x in X]),
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
import math
from pathlib import Path

import pytest

import gluonts
from gluonts.core.component import equals
from gluonts.core.serde import dump_code, load_code
from gluonts.dataset.artificial import constant_dataset
from gluonts.dataset.stat import ScaleHistogram  # noqa
from gluonts.dataset.stat import (
    DatasetStatistics,
    calculate_dataset_statistics,
)
from gluonts.evaluation import backtest_metrics, Evaluator
from gluonts.evaluation.backtest import BacktestInformation
from gluonts.model.trivial.mean import MeanEstimator

root = logging.getLogger()
root.setLevel(logging.DEBUG)


def make_estimator(freq, prediction_length):
    # noinspection PyTypeChecker
    return MeanEstimator(
        prediction_length=prediction_length, freq=freq, num_samples=5
    )


def test_forecast_parser():
    # verify that logged for estimator, datasets and metrics can be recovered
    # from their string representation

    dataset_info, train_ds, test_ds = constant_dataset()

    estimator = make_estimator(
        dataset_info.metadata.freq, dataset_info.prediction_length
    )
    assert repr(estimator) == repr(load_code(repr(estimator)))

    predictor = estimator.train(training_data=train_ds)

    stats = calculate_dataset_statistics(train_ds)
    assert stats == eval(
        repr(stats), globals(), {"gluonts": gluonts}
    )  # TODO: use load

    evaluator = Evaluator(quantiles=[0.1, 0.5, 0.9])
    agg_metrics, _ = backtest_metrics(test_ds, predictor, evaluator)

    # reset infinite metrics to 0 (otherwise the assertion below fails)
    for key, val in agg_metrics.items():
        if not math.isfinite(val):
            agg_metrics[key] = 0.0

    assert agg_metrics == load_code(dump_code(agg_metrics))


@pytest.mark.skip()
def test_benchmark(caplog):
    # makes sure that information logged can be reconstructed from previous
    # logs

    with caplog.at_level(logging.DEBUG):
        dataset_info, train_ds, test_ds = constant_dataset()

        estimator = make_estimator(
            dataset_info.metadata.freq, dataset_info.prediction_length
        )
        predictor = estimator.train(training_data=train_ds)
        evaluator = Evaluator(quantiles=[0.1, 0.5, 0.9])
        backtest_metrics(test_ds, predictor, evaluator)
        train_stats = calculate_dataset_statistics(train_ds)
        test_stats = calculate_dataset_statistics(test_ds)

    log_info = BacktestInformation.make_from_log_contents(caplog.text)

    assert train_stats == log_info.train_dataset_stats
    assert test_stats == log_info.test_dataset_stats
    assert equals(estimator, log_info.estimator)

    print(log_info)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import List, Literal, Optional, Union
import numpy as np
from tqdm.auto import tqdm
from tsbench.config import Config, MODEL_REGISTRY, ModelConfig, TrainConfig
from tsbench.config.dataset import get_dataset_config
from tsbench.config.model import get_model_config
from tsbench.constants import DEFAULT_DATA_PATH
from tsbench.evaluations.metrics import Metric, Performance
from .job import Job

ValidationMetric = Literal["train_loss", "val_loss", "val_ncrps"]


@dataclass
class ValidationScores:
    """
    Scores obtained during validation.
    """

    ncrps: float
    loss: float


@dataclass
class JobInfo:
    """
    The job info class aggregates all information available for a particular model configuration.
    It also provides the underlying training jobs. Lastly, it provides the indices of the models
    chosen from the training job to extract forecasts.
    """

    config: Config[ModelConfig]
    performance: Performance
    val_scores: Optional[ValidationScores]
    jobs: List[Job]
    model_indices: List[int]


# -------------------------------------------------------------------------------------------------


def extract_job_infos(
    training_jobs: List[Job],
    validation_metric: Optional[ValidationMetric],
    group_seeds: bool,
    data_path: Union[str, Path] = DEFAULT_DATA_PATH,
) -> List[JobInfo]:
    """
    Returns a list of the job information objects available for all training jobs provided.
    """
    # We group the jobs by hyperparameters, excluding the seed
    if group_seeds:
        grouped_jobs = defaultdict(list)
        for job in training_jobs:
            hypers = {
                "model": job.model,
                "dataset": job.dataset,
                **job.hyperparameters,
            }
            grouped_jobs[tuple(sorted(hypers.items()))].append(job)
        all_jobs = grouped_jobs.values()
    else:
        all_jobs = [[job] for job in training_jobs]

    # Then, we can instantiate the info objects by iterating over groups of jobs
    runs = []
    for jobs in tqdm(all_jobs):
        ref_job = jobs[0]
        model_name = ref_job.model
        base_hyperparams = {**ref_job.hyperparameters}

        # First, we reconstruct the training times
        if issubclass(MODEL_REGISTRY[model_name], TrainConfig):
            training_fractions = [1 / 81, 1 / 27] + [
                i / 9 for i in range(1, 10)
            ]
        else:
            training_fractions = [0]

        assert all(
            len(job.metrics) == len(training_fractions) for job in jobs
        ), "Job does not provide sufficiently many models."

        # Then, we iterate over the Hyperband training times
        if len(training_fractions) == 1:
            training_fraction_indices = [0]
        else:
            training_fraction_indices = [0, 1, 2, 4, 10]

        # Then, we iterate over all training times, construct the hyperparameters and collect
        # the performane metrics
        for i in training_fraction_indices:
            # Create the config object
            hyperparams = {
                **base_hyperparams,
                "training_fraction": training_fractions[i],
            }
            model_config = get_model_config(model_name, **hyperparams)
            config = Config(
                model_config, get_dataset_config(ref_job.dataset, data_path)
            )

            # Get the indices of the models that should be used to derive the performance
            if validation_metric is None or len(training_fractions) == 1:
                # If the model does not require training, or we don't look at the validation
                # performance, we just choose the current index
                choices = [i] * len(jobs)
            else:
                # Otherwise, we get the minimum value for the metric up to this point in time
                choices = [
                    np.argmin(
                        [
                            p["evaluation"][validation_metric]
                            for p in job.metrics
                        ][: i + 1]
                    ).item()
                    for job in jobs
                ]

            # Get the performances of the chosen models
            performances = [
                job.performances[choice] for choice, job in zip(choices, jobs)
            ]

            # And average the performance
            averaged_performance = Performance(
                **{
                    metric: Metric(
                        np.mean(
                            [getattr(p, metric).mean for p in performances]
                        ),
                        np.std(
                            [getattr(p, metric).mean for p in performances]
                        ),
                    )
                    for metric in Performance.metrics()
                }
            )

            # Get validation scores if available
            try:
                val_ncrps = np.mean(
                    [
                        job.metrics[c]["evaluation"]["val_ncrps"]
                        for (job, c) in zip(jobs, choices)
                    ]
                )
                val_loss = np.mean(
                    [
                        job.metrics[c]["evaluation"]["val_loss"]
                        for (job, c) in zip(jobs, choices)
                    ]
                ).item()
                val_scores = ValidationScores(val_ncrps, val_loss)
            except KeyError:
                val_scores = None

            # Initialize the info object
            runs.append(
                JobInfo(
                    config, averaged_performance, val_scores, jobs, choices
                )
            )

    return runs



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import time
from datetime import datetime, timezone
from pathlib import Path
import click
import sagemaker
from botocore.exceptions import ClientError
from cli.utils import generate_configurations, iterate_configurations
from tsbench.evaluations.aws import default_session
from tsbench.evaluations.aws.ecr import image_uri
from tsbench.evaluations.aws.framework import CustomFramework
from tsbench.evaluations.metrics.sagemaker import metric_definitions
from ._main import evaluations


@evaluations.command(short_help="Schedule evaluations on AWS Sagemaker.")
@click.option(
    "--config_path",
    required=True,
    help="The local path to the configuration file or directory. "
    "For directories, all YAML files in the directory and its subfolders are considered.",
)
@click.option(
    "--sagemaker_role",
    required=True,
    help="The IAM role to use for the AWS Sagemaker instance and for downloading data locally. "
    "This should be the full ARN.",
)
@click.option(
    "--experiment",
    default="tsbench",
    show_default=True,
    help="The name of the experiment which allows grouping training runs on AWS Sagemaker.",
)
@click.option(
    "--data_bucket",
    default="tsbench",
    show_default=True,
    help="The S3 bucket where the datasets have been uploaded to.",
)
@click.option(
    "--data_bucket_prefix",
    default="data",
    show_default=True,
    help="The prefix for the S3 bucket where datasets have been uploaded to.",
)
@click.option(
    "--output_bucket",
    default="tsbench",
    show_default=True,
    help="The S3 bucket where outputs (model parameters and forecasts) should be written to.",
)
@click.option(
    "--output_bucket_prefix",
    default="evaluations",
    show_default=True,
    help="The prefix for the S3 bucket where outputs are written to.",
)
@click.option(
    "--instance_type",
    default="ml.c5.2xlarge",
    show_default=True,
    help="The type of instance to use for running evaluations if not provided explicitly.",
)
@click.option(
    "--docker_image",
    default="tsbench:latest",
    show_default=True,
    help="The Docker image in your ECR registry to use for running the evaluation.",
)
@click.option(
    "--max_runtime",
    default=240,
    show_default=True,
    help="The maximum number of hours for which individual evaluations may run.",
)
@click.option(
    "--nskip",
    default=0,
    show_default=True,
    help="The number of configurations to skip. Useful if some set of evaluations failed.",
)
@click.option(
    "--local",
    default=False,
    show_default=True,
    help="Whether to run evaluations locally via Docker Compose instead of on AWS Sagemaker.",
)
def schedule(
    config_path: str,
    sagemaker_role: str,
    experiment: str,
    data_bucket: str,
    data_bucket_prefix: str,
    output_bucket: str,
    output_bucket_prefix: str,
    instance_type: str,
    docker_image: str,
    max_runtime: int,
    nskip: int,
    local: bool,
):
    """
    Schedules evaluations on AWS Sagemaker by running a grid search over the configurations
    provided in the given file(s). As AWS Sagemaker does not allow queuing jobs, this script is
    running as long as not all evaluation runs have been scheduled.
    """
    assert instance_type[:5] not in (
        "ml.p3",
        "ml.p2",
        "ml.g4",
    ), "Cannot schedule experiments on GPU instances."

    # First, setup Sagemaker connection
    boto_session = default_session()
    if local:
        sm_session = sagemaker.LocalSession(boto_session)
    else:
        sm_session = sagemaker.Session(boto_session)

    def job_factory() -> str:
        date_str = datetime.now(tz=timezone.utc).strftime(
            "%d-%m-%Y-%H-%M-%S-%f"
        )
        job_name = f"{experiment}-{date_str}"
        return job_name

    # Then, generate configs
    all_configurations = generate_configurations(Path(config_path))

    # Then, we can run the training, passing parameters as required
    for configuration in iterate_configurations(all_configurations, nskip):
        # Create the estimator
        estimator = CustomFramework(
            sagemaker_session=sm_session,
            role=sagemaker_role,
            tags=[
                {"Key": "Experiment", "Value": experiment},
            ],
            instance_type="local"
            if local
            else (
                configuration["__instance_type__"]
                if "__instance_type__" in configuration
                else instance_type
            ),
            instance_count=1,
            volume_size=30,
            max_run=max_runtime * 60 * 60,
            image_uri=image_uri(docker_image),
            source_dir=str(
                Path(os.path.realpath(__file__)).parent.parent.parent
            ),
            output_path=f"s3://{output_bucket}/{output_bucket_prefix}/{experiment}",
            entry_point="evaluate.py",
            debugger_hook_config=False,
            metric_definitions=metric_definitions(),
            hyperparameters={
                k: v
                for k, v in configuration.items()
                if not k.startswith("__")
            },
        )

        while True:
            # Try fitting the estimator
            try:
                estimator.fit(
                    job_name=job_factory(),
                    inputs={
                        configuration[
                            "dataset"
                        ]: f"s3://{data_bucket}/{data_bucket_prefix}/{configuration['dataset']}"
                    },
                    wait=False,
                )
                break
            except ClientError as err:
                print(f"+++ Scheduling failed: {err}")
                print("+++ Sleeping for 5 minutes.")
                time.sleep(300)

        print(f">>> Launched job: {estimator.latest_training_job.name}")  # type: ignore

    print(">>> Successfully scheduled all training jobs.")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Implementation taken and modified from
# https://github.com/White-Link/UnsupervisedScalableRepresentationLearningTimeSeries, which was created
# with the following license.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


# Implementation of causal CNNs partly taken and modified from
# https://github.com/locuslab/TCN/blob/master/TCN/tcn.py, originally created
# with the following license.

# MIT License

# Copyright (c) 2018 CMU Locus Lab

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import torch


class Chomp1d(torch.nn.Module):
    """Removes leading or trailing elements of a time series.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`, `L - s`) where `s`
    is the number of elements to remove.

    Args:
        chomp_size : Number of elements to remove.
        last : If True, removes the last elements in the time dimension,
            If False, removes the fist elements.
    """

    def __init__(self, chomp_size: int, last: bool = True):
        super(Chomp1d, self).__init__()
        self.chomp_size = chomp_size
        self.last = last

    def forward(self, x):
        if self.last:
            x_chomped = x[:, :, : -self.chomp_size]
        else:
            x_chomped = x[:, :, self.chomp_size :]

        return x_chomped


class TCNBlock(torch.nn.Module):
    """Temporal Convolutional Network block.

    Composed sequentially of two causal convolutions (with leaky ReLU activation functions),
    and a parallel residual connection.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`, `L`).

    Args:
        in_channels : Number of input channels.
        out_channels : Number of output channels.
        kernel_size : Kernel size of the applied non-residual convolutions.
        dilation : Dilation parameter of non-residual convolutions.
        bias : If True, adds a learnable bias to the convolutions.
        fwd_time : If True, the network "causal" direction is from past to future (forward),
            if False, the relation is from future to past (backward).
        final : If True, the last activation function is disabled.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
        fwd_time: bool = True,
        final: bool = False,
    ):

        super(TCNBlock, self).__init__()

        in_channels = int(in_channels)
        kernel_size = int(kernel_size)
        out_channels = int(out_channels)
        dilation = int(dilation)

        # Computes left padding so that the applied convolutions are causal
        padding = int((kernel_size - 1) * dilation)

        # First causal convolution
        conv1_pre = torch.nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )
        conv1 = torch.nn.utils.weight_norm(conv1_pre)

        # The truncation makes the convolution causal
        chomp1 = Chomp1d(chomp_size=padding, last=fwd_time)

        relu1 = torch.nn.LeakyReLU()

        # Second causal convolution
        conv2_pre = torch.nn.Conv1d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )
        conv2 = torch.nn.utils.weight_norm(conv2_pre)
        chomp2 = Chomp1d(padding)
        relu2 = torch.nn.LeakyReLU()

        # Causal network
        self.causal = torch.nn.Sequential(
            conv1, chomp1, relu1, conv2, chomp2, relu2
        )

        # Residual connection
        self.upordownsample = (
            torch.nn.Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
            )
            if in_channels != out_channels
            else None
        )

        # Final activation function
        self.activation = torch.nn.LeakyReLU() if final else None

    def forward(self, x):
        out_causal = self.causal(x)
        res = x if self.upordownsample is None else self.upordownsample(x)
        if self.activation is None:
            return out_causal + res
        else:
            return self.activation(out_causal + res)


class TCN(torch.nn.Module):
    """Temporal Convolutional Network.

    Composed of a sequence of causal convolution blocks.

    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C_out`, `L`).

    Args:
        in_channels : Number of input channels.
        out_channels : Number of output channels.
        kernel_size : Kernel size of the applied non-residual convolutions.
        channels : Number of channels processed in the network and of output
            channels.
        layers : Depth of the network.
        bias : If True, adds a learnable bias to the convolutions.
        fwd_time : If True the network is the relation relation if from past to future (forward),
            if False, the relation from future to past (backward).
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        channels: int,
        layers: int,
        bias: bool = True,
        fwd_time: bool = True,
    ):

        super(TCN, self).__init__()

        layers = int(layers)

        net_layers = []  # List of sequential TCN blocks
        dilation_size = 1  # Initial dilation size

        for i in range(layers):
            in_channels_block = in_channels if i == 0 else channels
            net_layers.append(
                TCNBlock(
                    in_channels=in_channels_block,
                    out_channels=channels,
                    kernel_size=kernel_size,
                    dilation=dilation_size,
                    bias=bias,
                    fwd_time=fwd_time,
                    final=False,
                )
            )
            dilation_size *= 2  # Doubles the dilation size at each step

        # Last layer
        net_layers.append(
            TCNBlock(
                in_channels=channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                dilation=dilation_size,
                bias=bias,
                fwd_time=fwd_time,
                final=True,
            )
        )

        self.network = torch.nn.Sequential(*net_layers)

    def forward(self, x):
        return self.network(x)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List
from .performance import Performance


def metric_definitions() -> List[Dict[str, str]]:
    """
    Returns the metric definitions to be used to collect performance measures from AWS Sagemaker.
    """
    # pylint: disable=no-member
    scalar_metrics = [
        {"Name": name, "Regex": _metric_regex(name)}
        for name in Performance.__dataclass_fields__  # type: ignore
    ]
    list_metrics = [
        {
            "Name": "train_loss",
            "Regex": f"'epoch_loss'={_FLOATING_POINT_REGEX}",
        },
        {
            "Name": "val_loss",
            "Regex": f"'validation_epoch_loss'={_FLOATING_POINT_REGEX}",
        },
        {
            "Name": "val_ncrps",
            "Regex": _metric_regex("val_ncrps"),
        },
    ]
    return scalar_metrics + list_metrics


# -------------------------------------------------------------------------------------------------

_FLOATING_POINT_REGEX = (
    r"(([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]\d+)?)|[Nn][Aa][Nn])"
)


def _metric_regex(target: str) -> str:
    return f"gluonts\\[{target}\\]: {_FLOATING_POINT_REGEX}"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np

from gluonts.model.util import (
    LinearInterpolation,
    ExponentialTailApproximation,
)


def test_linear_interpolation() -> None:
    tol = 1e-7
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]
    linear_interpolation = LinearInterpolation(x_coord, y_coord)
    x = 0.75
    exact = y_coord[1] + (x - x_coord[1]) * (y_coord[2] - y_coord[1]) / (
        x_coord[2] - x_coord[1]
    )
    assert np.all(np.abs(exact - linear_interpolation(x)) <= tol)


def test_exponential_left_tail_approximation() -> None:
    tol = 1e-5
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]
    x = 0.01
    beta_inv = np.array([0.55920144, 0.9320024, 1.24266987])
    exact = beta_inv * np.log(x / x_coord[1]) + y_coord[1]
    exp_tail_approximation = ExponentialTailApproximation(x_coord, y_coord)
    assert np.all(np.abs(exact - exp_tail_approximation.left(x)) <= tol)


def test_exponential_right_tail_approximation() -> None:
    tol = 1e-5
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]
    x = 0.99
    beta_inv = np.array([-0.4660012, -0.9320024, -1.30480336])
    exact = beta_inv * np.log((1 - x_coord[1]) / (1 - x)) + y_coord[1]
    exp_tail_approximation = ExponentialTailApproximation(x_coord, y_coord)
    assert np.all(np.abs(exact - exp_tail_approximation.right(x)) <= tol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# GluonTS documentation build configuration file, created by
# sphinx-quickstart on Wed Dec 19 15:29:41 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
import sys
import os, subprocess
import shlex
import recommonmark

from recommonmark.parser import CommonMarkParser
from recommonmark.transform import AutoStructify

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
curr_path = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
sys.path.insert(0, os.path.join(curr_path, ".."))


# -- General configuration ------------------------------------------------

# Version information.
from pkg_resources import get_distribution

release = get_distribution("gluonts").version
version = ".".join(release.split(".")[:2])  # strip major.minor version only

# General information about the project.
project = "GluonTS"
copyright = "2019, Amazon"
author = "Amazon"
github_doc_root = "http://gluon-ts.mxnet.io/{}/".format(str(version))

# add markdown parser
CommonMarkParser.github_doc_root = github_doc_root
source_parsers = {".md": CommonMarkParser}

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

nbsphinx_kernel_name = "python3"
nbsphinx_allow_errors = True
nbsphinx_timeout = 1200
html_sourcelink_suffix = ""

html_context = {
    "display_github": True,
    "github_user": "awslabs",
    "github_repo": "gluon-ts",
    "github_version": "master",
    "conf_py_path": "/docs/",
    "last_updated": False,
    "commit": True,
}

nbsphinx_prolog = """
{% set paths = env.docname.split('/') %}

.. only:: html

    :download:`[Download] <{{ "../%s.zip"|format(paths[1]) }}>`
"""
# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = [".rst", ".ipynb", ".md"]

# generate autosummary even if no references
autosummary_generate = True

# The master toctree document.
master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ""
# The full version, including alpha/beta/rc tags.
release = ""

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/gluon-logo.svg"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/gluon.ico"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "**.ipynb_checkpoints", "examples/*/*/**.rst"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "mxtheme"
html_theme_path = ["mxtheme"]
html_theme_options = {
    "primary_color": "blue",
    "accent_color": "deep_orange",
    "header_links": [
        ("Install", "install", False, ""),
        ("API", "api/index", False, ""),
        ("Community", "community/index", False, ""),
        ("Contribute", "community/contribute", False, ""),
        (
            "GitHub",
            "https://github.com/awslabs/gluon-ts/",
            True,
            "fab fa-github",
        ),
    ],
    # custom layout
    "fixed_drawer": True,
    "fixed_header": True,
    "header_waterfall": True,
    "header_scroll": True,
    # Render footer (Default: True)
    "show_footer": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Output file base name for HTML help builder.
htmlhelp_basename = "GluonTSdoc"


# -- Options for LaTeX output ---------------------------------------------

# latex_elements = {
# }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#     (master_doc, 'GluonTS.tex', 'GluonTS Documentation', 'Amazon', 'manual')
# ]

intersphinx_mapping = {
    "python": (
        "https://docs.python.org/{.major}".format(sys.version_info),
        None,
    ),
    "mxnet": ("https://mxnet.apache.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("http://matplotlib.org/", None),
}

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "gluonts", "GluonTS Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "GluonTS",
        "GluonTS Documentation",
        author,
        "GluonTS",
        "One line description of project.",
        "Miscellaneous",
    )
]


def setup(app):
    import mxtheme

    app.add_directive("card", mxtheme.CardDirective)

    app.add_config_value(
        "recommonmark_config",
        {
            "url_resolver": lambda url: github_doc_root + url,
            "auto_doc_ref": True,
        },
        True,
    )
    app.add_transform(AutoStructify)
    app.add_js_file("google_analytics.js")


# Napoleon settings
napoleon_use_ivar = True

# linkcheck settings
import multiprocessing

linkcheck_ignore = [r"http[s]://apache-mxnet.s3*"]
linkcheck_retries = 3
linkcheck_workers = int(multiprocessing.cpu_count() / 2)

intersphinx_mapping = {
    "mxnet": ("https://mxnet.apache.org/api/python/docs", None)
}



from typing import List, Tuple

import numpy as np
import pytest
import torch
import torch.nn as nn
from torch.distributions import (
    StudentT,
    Beta,
    NegativeBinomial,
    LowRankMultivariateNormal,
    MultivariateNormal,
    Independent,
    Normal,
)
from torch.nn.utils import clip_grad_norm_
from torch.optim import SGD
from torch.utils.data import TensorDataset, DataLoader

from pts.modules import (
    DistributionOutput,
    StudentTOutput,
    BetaOutput,
    NegativeBinomialOutput,
    LowRankMultivariateNormalOutput,
    MultivariateNormalOutput,
    NormalOutput,
)

NUM_SAMPLES = 2000
BATCH_SIZE = 32
TOL = 0.3
START_TOL_MULTIPLE = 1


def inv_softplus(y: np.ndarray) -> np.ndarray:
    return np.log(np.exp(y) - 1)


def maximum_likelihood_estimate_sgd(
    distr_output: DistributionOutput,
    samples: torch.Tensor,
    init_biases: List[np.ndarray] = None,
    num_epochs: int = 5,
    learning_rate: float = 1e-2,
):
    arg_proj = distr_output.get_args_proj(in_features=1)

    if init_biases is not None:
        for param, bias in zip(arg_proj.proj, init_biases):
            nn.init.constant_(param.bias, bias)

    dummy_data = torch.ones((len(samples), 1))

    dataset = TensorDataset(dummy_data, samples)
    train_data = DataLoader(dataset, batch_size=BATCH_SIZE, shuffle=True)

    optimizer = SGD(arg_proj.parameters(), lr=learning_rate)

    for e in range(num_epochs):
        cumulative_loss = 0
        num_batches = 0

        for i, (data, sample_label) in enumerate(train_data):
            optimizer.zero_grad()
            distr_args = arg_proj(data)
            distr = distr_output.distribution(distr_args)
            loss = -distr.log_prob(sample_label).mean()
            loss.backward()
            clip_grad_norm_(arg_proj.parameters(), 10.0)
            optimizer.step()

            num_batches += 1
            cumulative_loss += loss.item()
        print("Epoch %s, loss: %s" % (e, cumulative_loss / num_batches))

    if len(distr_args[0].shape) == 1:
        return [
            param.detach().numpy() for param in arg_proj(torch.ones((1, 1)))
        ]

    return [
        param[0].detach().numpy() for param in arg_proj(torch.ones((1, 1)))
    ]


@pytest.mark.parametrize("concentration1, concentration0", [(3.75, 1.25)])
def test_beta_likelihood(concentration1: float, concentration0: float) -> None:
    """
    Test to check that maximizing the likelihood recovers the parameters
    """

    # generate samples
    concentration1s = torch.zeros((NUM_SAMPLES,)) + concentration1
    concentration0s = torch.zeros((NUM_SAMPLES,)) + concentration0

    distr = Beta(concentration1s, concentration0s)
    samples = distr.sample()

    init_biases = [
        inv_softplus(
            concentration1 - START_TOL_MULTIPLE * TOL * concentration1
        ),
        inv_softplus(
            concentration0 - START_TOL_MULTIPLE * TOL * concentration0
        ),
    ]

    concentration1_hat, concentration0_hat = maximum_likelihood_estimate_sgd(
        BetaOutput(),
        samples,
        init_biases=init_biases,
        learning_rate=0.05,
        num_epochs=10,
    )

    print(
        "concentration1:",
        concentration1_hat,
        "concentration0:",
        concentration0_hat,
    )
    assert (
        np.abs(concentration1_hat - concentration1) < TOL * concentration1
    ), f"concentration1 did not match: concentration1 = {concentration1}, concentration1_hat = {concentration1_hat}"
    assert (
        np.abs(concentration0_hat - concentration0) < TOL * concentration0
    ), f"concentration0 did not match: concentration0 = {concentration0}, concentration0_hat = {concentration0_hat}"


@pytest.mark.parametrize("total_count_logit", [(2.5, 0.7)])
def test_neg_binomial(total_count_logit: Tuple[float, float]) -> None:
    """
    Test to check that maximizing the likelihood recovers the parameters
    """
    # test instance
    total_count, logit = total_count_logit

    # generate samples
    total_counts = torch.zeros((NUM_SAMPLES,)) + total_count
    logits = torch.zeros((NUM_SAMPLES,)) + logit

    neg_bin_distr = NegativeBinomial(total_count=total_counts, logits=logits)
    samples = neg_bin_distr.sample()

    init_biases = [
        inv_softplus(total_count - START_TOL_MULTIPLE * TOL * total_count),
        logit - START_TOL_MULTIPLE * TOL * logit,
    ]

    total_count_hat, logit_hat = maximum_likelihood_estimate_sgd(
        NegativeBinomialOutput(),
        samples,
        init_biases=init_biases,
        num_epochs=15,
    )

    assert (
        np.abs(total_count_hat - total_count) < TOL * total_count
    ), f"total_count did not match: total_count = {total_count}, total_count_hat = {total_count_hat}"
    assert (
        np.abs(logit_hat - logit) < TOL * logit_hat
    ), f"logit did not match: logit = {logit}, logit_hat = {logit_hat}"


@pytest.mark.parametrize("df, loc, scale,", [(6.0, 2.3, 0.7)])
def test_studentT_likelihood(df: float, loc: float, scale: float):

    dfs = torch.zeros((NUM_SAMPLES,)) + df
    locs = torch.zeros((NUM_SAMPLES,)) + loc
    scales = torch.zeros((NUM_SAMPLES,)) + scale

    distr = StudentT(df=dfs, loc=locs, scale=scales)
    samples = distr.sample()

    init_bias = [
        inv_softplus(df - 2),
        loc - START_TOL_MULTIPLE * TOL * loc,
        inv_softplus(scale - START_TOL_MULTIPLE * TOL * scale),
    ]

    df_hat, loc_hat, scale_hat = maximum_likelihood_estimate_sgd(
        StudentTOutput(),
        samples,
        init_biases=init_bias,
        num_epochs=15,
        learning_rate=1e-3,
    )

    assert (
        np.abs(df_hat - df) < TOL * df
    ), f"df did not match: df = {df}, df_hat = {df_hat}"
    assert (
        np.abs(loc_hat - loc) < TOL * loc
    ), f"loc did not match: loc = {loc}, loc_hat = {loc_hat}"
    assert (
        np.abs(scale_hat - scale) < TOL * scale
    ), f"scale did not match: scale = {scale}, scale_hat = {scale_hat}"


def test_independent_normal() -> None:
    num_samples = 2000
    dim = 4

    loc = np.arange(0, dim) / float(dim)
    diag = np.arange(dim) / dim + 0.5
    Sigma = diag**2

    distr = Independent(
        Normal(loc=torch.Tensor(loc), scale=torch.Tensor(diag)), 1
    )

    assert np.allclose(
        distr.variance.numpy(), Sigma, atol=0.1, rtol=0.1
    ), f"did not match: sigma = {Sigma}, sigma_hat = {distr.variance.numpy()}"

    samples = distr.sample((num_samples,))

    loc_hat, diag_hat = maximum_likelihood_estimate_sgd(
        NormalOutput(dim=dim),
        samples,
        learning_rate=0.01,
        num_epochs=10,
    )

    distr = Independent(
        Normal(loc=torch.Tensor(loc_hat), scale=torch.Tensor(diag_hat)), 1
    )

    Sigma_hat = distr.variance.numpy()

    assert np.allclose(
        loc_hat, loc, atol=0.2, rtol=0.1
    ), f"mu did not match: loc = {loc}, loc_hat = {loc_hat}"

    assert np.allclose(
        Sigma_hat, Sigma, atol=0.1, rtol=0.1
    ), f"sigma did not match: sigma = {Sigma}, sigma_hat = {Sigma_hat}"


def test_lowrank_multivariate_normal() -> None:
    num_samples = 2000
    dim = 4
    rank = 3

    loc = np.arange(0, dim) / float(dim)
    cov_diag = np.eye(dim) * (np.arange(dim) / dim + 0.5)
    cov_factor = np.sqrt(np.ones((dim, rank)) * 0.2)
    Sigma = cov_factor @ cov_factor.T + cov_diag

    distr = LowRankMultivariateNormal(
        loc=torch.Tensor(loc.copy()),
        cov_diag=torch.Tensor(np.diag(cov_diag).copy()),
        cov_factor=torch.Tensor(cov_factor.copy()),
    )

    assert np.allclose(
        distr.covariance_matrix.numpy(), Sigma, atol=0.1, rtol=0.1
    ), f"did not match: sigma = {Sigma}, sigma_hat = {distr.covariance_matrix.numpy()}"

    samples = distr.sample((num_samples,))

    loc_hat, cov_factor_hat, cov_diag_hat = maximum_likelihood_estimate_sgd(
        LowRankMultivariateNormalOutput(
            dim=dim, rank=rank, sigma_init=0.2, sigma_minimum=0.0
        ),
        samples,
        learning_rate=0.01,
        num_epochs=10,
    )

    distr = LowRankMultivariateNormal(
        loc=torch.Tensor(loc_hat),
        cov_diag=torch.Tensor(cov_diag_hat),
        cov_factor=torch.Tensor(cov_factor_hat),
    )

    Sigma_hat = distr.covariance_matrix.numpy()

    assert np.allclose(
        loc_hat, loc, atol=0.2, rtol=0.1
    ), f"mu did not match: loc = {loc}, loc_hat = {loc_hat}"

    assert np.allclose(
        Sigma_hat, Sigma, atol=0.1, rtol=0.1
    ), f"sigma did not match: sigma = {Sigma}, sigma_hat = {Sigma_hat}"


def test_multivariate_normal() -> None:
    num_samples = 2000
    dim = 2

    mu = np.arange(0, dim) / float(dim)

    L_diag = np.ones((dim,))
    L_low = 0.1 * np.ones((dim, dim)) * np.tri(dim, k=-1)
    L = np.diag(L_diag) + L_low
    Sigma = L.dot(L.transpose())

    distr = MultivariateNormal(
        loc=torch.Tensor(mu), scale_tril=torch.Tensor(L)
    )

    samples = distr.sample((num_samples,))

    mu_hat, L_hat = maximum_likelihood_estimate_sgd(
        MultivariateNormalOutput(dim=dim),
        samples,
        init_biases=None,  # todo we would need to rework biases a bit to use it in the multivariate case
        learning_rate=0.01,
        num_epochs=10,
    )

    distr = MultivariateNormal(
        loc=torch.tensor(mu_hat), scale_tril=torch.tensor(L_hat)
    )

    Sigma_hat = distr.covariance_matrix.numpy()

    assert np.allclose(
        mu_hat, mu, atol=0.1, rtol=0.1
    ), f"mu did not match: mu = {mu}, mu_hat = {mu_hat}"
    assert np.allclose(
        Sigma_hat, Sigma, atol=0.1, rtol=0.1
    ), f"Sigma did not match: sigma = {Sigma}, sigma_hat = {Sigma_hat}"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import re
from functools import lru_cache
from typing import Tuple


def get_granularity(freq_str: str) -> Tuple[int, str]:
    """
    Splits a frequency string such as "7D" into the multiple 7 and the base
    granularity "D".

    Parameters
    ----------

    freq_str
        Frequency string of the form [multiple][granularity] such as "12H", "5min", "1D" etc.
    """
    freq_regex = r"\s*((\d+)?)\s*([^\d]\w*)"
    m = re.match(freq_regex, freq_str)
    assert m is not None, "Cannot parse frequency string: %s" % freq_str
    groups = m.groups()
    multiple = int(groups[1]) if groups[1] is not None else 1
    granularity = groups[2]
    return multiple, granularity


@lru_cache()
def get_seasonality(freq: str) -> int:
    """
    Returns the default seasonality for a given freq str. E.g. for

      2H -> 12

    """
    match = re.match(r"(\d*)(\w+)", freq)
    assert match, "Cannot match freq regex"
    mult, base_freq = match.groups()
    multiple = int(mult) if mult else 1

    seasonalities = {"H": 24, "D": 1, "W": 1, "M": 12, "B": 5}
    if base_freq in seasonalities:
        seasonality = seasonalities[base_freq]
    else:
        seasonality = 1
    if seasonality % multiple != 0:
        # logging.warning(
        #     f"multiple {multiple} does not divide base "
        #     f"seasonality {seasonality}."
        #     f"Falling back to seasonality 1"
        # )
        return 1
    return seasonality // multiple



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import io
import logging
import multiprocessing as mp
import pickle
import sys
from functools import partial
from multiprocessing.reduction import ForkingPickler
from typing import Callable, Iterable, Optional

from pydantic import BaseModel

from gluonts.dataset.common import DataBatch, Dataset
from gluonts.dataset.util import MPWorkerInfo
from gluonts.itertools import Cyclic, IterableSlice, PseudoShuffled, batcher
from gluonts.transform import AdhocTransform, Identity, Transformation

logger = logging.getLogger(__name__)


def win32_guard(cls, num_workers):
    if num_workers is None:
        return None

    assert num_workers > 0, "num_workers can't be negative"

    if sys.platform == "win32":
        logger.warning(
            "Multiprocessing is not supported on Windows, "
            "num_workers will be set to None."
        )
        return None

    return num_workers


def _encode(value):
    buf = io.BytesIO()
    ForkingPickler(buf, pickle.HIGHEST_PROTOCOL).dump(value)
    return buf.getvalue()


def worker_fn(
    worker_id: int,
    dataset,
    num_workers: int,
    result_queue: mp.Queue,
):
    MPWorkerInfo.set_worker_info(
        num_workers=num_workers,
        worker_id=worker_id,
    )

    for raw in map(_encode, dataset):
        try:
            result_queue.put(raw)
        except (EOFError, BrokenPipeError):
            return


DataLoader = Iterable[DataBatch]


class MultiProcessLoader(DataLoader):
    def __init__(
        self,
        dataset: Dataset,
        num_workers: int,
        max_queue_size: Optional[int],
        decode_fn: Callable = lambda x: x,
    ):
        assert num_workers >= 1

        if max_queue_size is None:
            max_queue_size = 5 * num_workers
        else:
            assert max_queue_size >= num_workers

        self.decode_fn = decode_fn
        self.result_queue = mp.Manager().Queue(maxsize=max_queue_size)

        create_worker = partial(
            mp.Process,
            target=worker_fn,
            kwargs={
                "dataset": dataset,
                "result_queue": self.result_queue,
                "num_workers": num_workers,
            },
        )

        self.processes = [
            create_worker(args=[worker_id]) for worker_id in range(num_workers)
        ]
        for process in self.processes:
            process.start()

    def _has_values(self):
        alive = any(proc.is_alive() for proc in self.processes)
        return alive or not self.result_queue.empty()

    def __iter__(self):
        while self._has_values():
            yield self._get()

        self._terminate()

    def _get(self):
        # TODO make timeout configurable
        raw = self.result_queue.get(timeout=120)
        return self.decode_fn(pickle.loads(raw))

    def _terminate(self):
        for process in self.processes:
            process.terminate()


# TODO: the following are for backward compatibility
# and could eventually be removed


class Batch(Transformation, BaseModel):
    batch_size: int

    def __call__(self, data, is_train):
        yield from batcher(data, self.batch_size)


def TrainDataLoader(
    dataset: Dataset,
    *,
    transform: Transformation = Identity(),
    batch_size: int,
    stack_fn: Callable,
    num_batches_per_epoch: Optional[int] = None,
    num_prefetch: Optional[int] = None,
    num_workers: Optional[int] = None,
    shuffle_buffer_length: Optional[int] = None,
    decode_fn: Callable = lambda x: x,
):
    """Construct an iterator of batches for training purposes.

    This function wraps around ``DataLoader`` to offer training-specific
    behaviour and options, as follows:

        1. The provided dataset is iterated cyclically, so that one can go over
        it multiple times in a single epoch. 2. A transformation must be
        provided, that is lazily applied as the dataset is being iterated;
        this is useful e.g. to slice random instances of fixed length out of
        each time series in the dataset. 3. The resulting batches can be
        iterated in a pseudo-shuffled order.

    The returned object is a stateful iterator, whose length is either
    ``num_batches_per_epoch`` (if not ``None``) or infinite (otherwise).

    Parameters
    ----------
    dataset
        Data to iterate over.
    transform
        Transformation to be lazily applied as data is being iterated.
        The transformation is applied in "training mode" (``is_train=True``).
    batch_size
        Number of entries to include in a batch.
    stack_fn
        Function to use to stack data entries into batches.
        This can be used to set a specific array type or computing device
        the arrays should end up onto (CPU, GPU).
    num_batches_per_epoch
        Length of the iterator. If ``None``, then the iterator is endless.
    num_workers
        Number of worker processes to use. Default: None.
    num_prefetch
        Sets the length of the queue of batches being produced by worker
        processes. (Only meaningful when ``num_workers is not None``).
    shuffle_buffer_length
        Size of the buffer used for shuffling. Default: None, in which case no
        shuffling occurs.
    decode_fn
        A function called on each batch after it's been taken out of the queue.
        (Only meaningful when ``num_workers is not None``).

    Returns
    -------
    Iterator[DataBatch]
        An iterator of batches.
    """
    dataset: Dataset = Cyclic(dataset)

    if shuffle_buffer_length:
        dataset = PseudoShuffled(dataset, shuffle_buffer_length)

    transform += Batch(batch_size=batch_size) + AdhocTransform(stack_fn)
    transformed_dataset = transform.apply(dataset, is_train=True)

    if num_workers is not None:
        loader = MultiProcessLoader(
            transformed_dataset,
            decode_fn=decode_fn,
            num_workers=num_workers,
            max_queue_size=num_prefetch,
        )
        batches = iter(loader)
    else:
        batches = iter(transformed_dataset)

    if num_batches_per_epoch is None:
        return batches
    else:
        return IterableSlice(batches, num_batches_per_epoch)


def ValidationDataLoader(
    dataset: Dataset,
    *,
    transform: Transformation = Identity(),
    batch_size: int,
    stack_fn: Callable,
):
    """Construct an iterator of batches for validation purposes.

    Parameters
    ----------
    dataset
        Data to iterate over.
    transform
        Transformation to be lazily applied as data is being iterated.
        The transformation is applied in "training mode" (``is_train=True``).
    batch_size
        Number of entries to include in a batch.
    stack_fn
        Function to use to stack data entries into batches.
        This can be used to set a specific array type or computing device
        the arrays should end up onto (CPU, GPU).

    Returns
    -------
    Iterable[DataBatch]
        An iterable sequence of batches.
    """

    transform += Batch(batch_size=batch_size) + AdhocTransform(stack_fn)
    return transform.apply(dataset, is_train=True)


def InferenceDataLoader(
    dataset: Dataset,
    *,
    transform: Transformation = Identity(),
    batch_size: int,
    stack_fn: Callable,
):
    """Construct an iterator of batches for inference purposes.

    Parameters
    ----------
    dataset
        Data to iterate over.
    transform
        Transformation to be lazily applied as data is being iterated.
        The transformation is applied in "inference mode" (``is_train=False``).
    batch_size
        Number of entries to include in a batch.
    stack_fn
        Function to use to stack data entries into batches.
        This can be used to set a specific array type or computing device
        the arrays should end up onto (CPU, GPU).

    Returns
    -------
    Iterable[DataBatch]
        An iterable sequence of batches.
    """
    transform += Batch(batch_size=batch_size) + AdhocTransform(stack_fn)
    return transform.apply(dataset, is_train=False)



from .holiday import (
    SPECIAL_DATE_FEATURES,
    SpecialDateFeatureSet,
    CustomDateFeatureSet,
    CustomHolidayFeatureSet,
    squared_exponential_kernel,
    exponential_kernel,
)
from .lag import get_lags_for_frequency, get_fourier_lags_for_frequency
from .time_feature import (
    DayOfMonth,
    DayOfWeek,
    DayOfYear,
    HourOfDay,
    MinuteOfHour,
    MonthOfYear,
    TimeFeature,
    WeekOfYear,
    FourierDateFeatures,
    time_features_from_frequency_str,
    fourier_time_features_from_frequency_str,
)
from .utils import get_granularity, get_seasonality



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any, Dict, Iterator, List, Optional, Union
import gridfs
import pymongo
from .experiment import SacredExperiment


class SacredMongoClient:
    """
    The Sacred Mongo client allows to retrieve results from experiments tracked with Sacred from
    MongoDB.
    """

    base_query: Dict[str, Any]

    def __init__(
        self,
        experiment: Union[str, List[str]],
        host: str = "localhost",
        port: int = 27017,
        database: str = "sacred",
        only_completed: bool = True,
    ):
        """
        Args:
            experiment: The name(s) of the experiment(s) for which to query individual runs.
            only_completed: Whether only completed jobs should be returned.
            host: The hostname of the server where the MongoDB is running.
            port: The port where the MongoDB is reachable.
            database: The name of the database where Sacred experiments are stored.
        """
        # Initialize Mongo client
        client = pymongo.MongoClient(host, port)
        self.db = client[database]
        self.fs = gridfs.GridFS(self.db)

        # Initialize base query
        if isinstance(experiment, list):
            self.base_query = {"config.experiment": {"$in": experiment}}
        else:
            self.base_query = {"config.experiment": experiment}

        if only_completed:
            self.base_query.update({"status": "COMPLETED"})

    def query_one(
        self, config: Dict[str, Any], others: Optional[Dict[str, Any]] = None
    ) -> SacredExperiment:
        """
        Queries the experiments associated with the experiment set and returns the one matching
        the query.

        Args:
            config: The query items that apply to the configuration.
            others: Optional additional query options.

        Returns:
            The sacred experiment describing the experiment.
        """
        full_query = {
            **self.base_query,
            **{f"config.{k}": v for k, v in config.items()},
            **(others or {}),
        }
        assert (
            self.db.runs.count_documents(full_query) == 1
        ), "Query does not return a single experiment."

        info = self.db.runs.find_one(full_query)
        assert info is not None

        return SacredExperiment(info, self.db, self.fs)

    def query(
        self, config: Dict[str, Any], others: Optional[Dict[str, Any]] = None
    ) -> List[SacredExperiment]:
        """
        Queries the experiments associated with the experiment set and returns all matching the
        query.

        Args:
            config: The query items that apply to the configuration.
            others: Optional additional query options.

        Returns:
            The sacred experiments found via the query.
        """
        full_query = {
            **self.base_query,
            **{f"config.{k}": v for k, v in config.items()},
            **(others or {}),
        }
        infos = self.db.runs.find(full_query)
        return [SacredExperiment(info, self.db, self.fs) for info in infos]

    def __iter__(self) -> Iterator[SacredExperiment]:
        for info in self.db.runs.find(self.base_query):
            yield SacredExperiment(info, self.db, self.fs)

    def __len__(self) -> int:
        return self.db.runs.count_documents(self.base_query)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from functools import lru_cache
from typing import Callable, List, cast

import numpy as np
import pandas as pd
from pandas.tseries.offsets import Tick

from .common import DataEntry


class ProcessStartField:
    def __init__(self, name: str, freq: str) -> None:
        self.name = name
        self.freq = freq

    def __call__(self, data: DataEntry) -> DataEntry:
        try:
            value = ProcessStartField.process(data[self.name], self.freq)
        except (TypeError, ValueError) as e:
            raise Exception(
                f'Error "{e}" occurred when reading field "{self.name}"'
            )

        data[self.name] = value

        return data

    @staticmethod
    @lru_cache(maxsize=10000)
    def process(string: str, freq: str) -> pd.Timestamp:
        timestamp = pd.Timestamp(string, freq=freq)

        # operate on time information (days, hours, minute, second)
        if isinstance(timestamp.freq, Tick):
            return pd.Timestamp(
                timestamp.floor(timestamp.freq), timestamp.freq
            )

        # since we are only interested in the data piece, we normalize the
        # time information
        timestamp = timestamp.replace(
            hour=0, minute=0, second=0, microsecond=0, nanosecond=0
        )

        return timestamp.freq.rollforward(timestamp)


class ProcessTimeSeriesField:
    def __init__(
        self, name, is_required: bool, is_static: bool, is_cat: bool
    ) -> None:
        self.name = name
        self.is_required = is_required
        self.req_ndim = 1 if is_static else 2
        self.dtype = np.int64 if is_cat else np.float32

    def __call__(self, data: DataEntry) -> DataEntry:
        value = data.get(self.name, None)

        if value is not None:
            value = np.asarray(value, dtype=self.dtype)
            dim_diff = self.req_ndim - value.ndim
            if dim_diff == 1:
                value = np.expand_dims(a=value, axis=0)
            elif dim_diff != 0:
                raise Exception(
                    f"JSON array has bad shape - expected {self.req_ndim} dimensions got {dim_diff}"
                )

            data[self.name] = value
            return data
        elif not self.is_required:
            return data
        else:
            raise Exception(
                f"JSON object is missing a required field `{self.name}`"
            )


class ProcessDataEntry:
    def __init__(self, freq: str, one_dim_target: bool = True) -> None:
        self.trans = cast(
            List[Callable[[DataEntry], DataEntry]],
            [
                ProcessStartField("start", freq=freq),
                ProcessTimeSeriesField(
                    "target",
                    is_required=True,
                    is_cat=False,
                    is_static=one_dim_target,
                ),
                ProcessTimeSeriesField(
                    "feat_dynamic_cat",
                    is_required=False,
                    is_cat=True,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    "feat_dynamic_real",
                    is_required=False,
                    is_cat=False,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    "feat_static_cat",
                    is_required=False,
                    is_cat=True,
                    is_static=True,
                ),
                ProcessTimeSeriesField(
                    "feat_static_real",
                    is_required=False,
                    is_cat=False,
                    is_static=True,
                ),
            ],
        )

    def __call__(self, data: DataEntry) -> DataEntry:
        for t in self.trans:
            data = t(data)
        return data



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math
import multiprocessing as mp
import os
from multiprocessing.pool import ThreadPool
from queue import Empty
from typing import Any, Callable, cast, List, TypeVar
import psutil
from tqdm.auto import tqdm

T = TypeVar("T")
U = TypeVar("U")


_GLOBAL_DATA_CACHE = None


def num_fitting_processes(
    cpus_per_process: float = 2, memory_per_process: float = 16
) -> int:
    """
    Returns the number of processes that can be fitted onto the machine when using a particular
    number of CPUs and a particular amount of memory for every process.

    Args:
        cpus_per_process: The number of CPUs that every process requires.
        memory_per_process: The memory in GiB that every process needs.

    Returns:
        The number of processes to use.
    """
    num_processes_cpu = math.floor(
        cast(int, os.cpu_count()) / cpus_per_process
    )

    available_gib = psutil.virtual_memory().total / (1024**3)
    num_processes_memory = math.floor(available_gib / memory_per_process)

    return min(num_processes_cpu, num_processes_memory)


def run_parallel(
    execute: Callable[[T], U], data: List[T], num_processes: int
) -> List[U]:
    """
    Runs a function on multiple processes, parallelizing computations for the provided data.

    Args:
        execute: The function to run in each process.
        data: The data items to execute the function for.
        num_processes: The number of processes to parallelize over.

    Returns:
        The outputs of the function calls, ordered in the same way as the data.
    """

    def factory(_i: int) -> mp.Process:
        process = mp.Process(target=_worker, args=(execute, inputs, outputs))
        process.start()
        return process

    # We share data via the global data cache such that we do not need to pass too much data over
    # queues -- forking allows to read the same data from multiple processes without issues.
    global _GLOBAL_DATA_CACHE  # pylint: disable=global-statement
    _GLOBAL_DATA_CACHE = data  # type: ignore

    # Initialize queues and put all items into input queue. Also, put as many "done" items into the
    # queue as we have processes
    inputs = mp.Queue()
    outputs = mp.Queue()
    for i in range(len(data)):
        inputs.put(i)

    # Create the processes in a thread pool to speed up creation
    with ThreadPool(num_processes) as p:
        processes = p.map_async(factory, range(num_processes))

        # Parallelize execution -- keep this inside the with statement so processes keep
        # getting spawned
        result = [None] * len(data)
        with tqdm(total=len(data)) as progress:
            progress.set_postfix({"num_processes": num_processes})
            for _ in range(len(data)):
                i, output = outputs.get()
                result[i] = output
                progress.update()

        for p in processes.get():
            p.kill()

    # cleanup
    inputs.close()
    outputs.close()
    return cast(List[U], result)


def _worker(
    execute: Callable[[Any], Any],
    inputs: mp.Queue,  # type: ignore
    outputs: mp.Queue,  # type: ignore
) -> None:
    while True:
        # Timeout is needed to shutdown workers if no tasks are available anymore.
        try:
            i = inputs.get(timeout=10)
        except Empty:
            return
        data = _GLOBAL_DATA_CACHE[i]  # type: ignore
        output = execute(data)
        outputs.put((i, output))



from abc import ABC, abstractmethod
from typing import Tuple

import torch
import torch.nn as nn


class Scaler(ABC, nn.Module):
    def __init__(self, keepdim: bool = False, time_first: bool = True):
        super().__init__()
        self.keepdim = keepdim
        self.time_first = time_first

    @abstractmethod
    def compute_scale(
        self, data: torch.Tensor, observed_indicator: torch.Tensor
    ) -> torch.Tensor:
        pass

    def forward(
        self, data: torch.Tensor, observed_indicator: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Parameters
        ----------
        data
            tensor of shape (N, T, C) if ``time_first == True`` or (N, C, T)
            if ``time_first == False`` containing the data to be scaled

        observed_indicator
            observed_indicator: binary tensor with the same shape as
            ``data``, that has 1 in correspondence of observed data points,
            and 0 in correspondence of missing data points.

        Returns
        -------
        Tensor
            Tensor containing the "scaled" data, shape: (N, T, C) or (N, C, T).
        Tensor
            Tensor containing the scale, of shape (N, C) if ``keepdim == False``,
            and shape (N, 1, C) or (N, C, 1) if ``keepdim == True``.
        """

        scale = self.compute_scale(data, observed_indicator)

        if self.time_first:
            dim = 1
        else:
            dim = 2
        if self.keepdim:
            scale = scale.unsqueeze(dim=dim)
            return data / scale, scale
        else:
            return data / scale.unsqueeze(dim=dim), scale


class MeanScaler(Scaler):
    """
    The ``MeanScaler`` computes a per-item scale according to the average
    absolute value over time of each item. The average is computed only among
    the observed values in the data tensor, as indicated by the second
    argument. Items with no observed data are assigned a scale based on the
    global average.

    Parameters
    ----------
    minimum_scale
        default scale that is used if the time series has only zeros.
    """

    def __init__(self, minimum_scale: float = 1e-10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_buffer("minimum_scale", torch.tensor(minimum_scale))

    def compute_scale(
        self, data: torch.Tensor, observed_indicator: torch.Tensor
    ) -> torch.Tensor:

        if self.time_first:
            dim = 1
        else:
            dim = 2

        # these will have shape (N, C)
        num_observed = observed_indicator.sum(dim=dim)
        sum_observed = (data.abs() * observed_indicator).sum(dim=dim)

        # first compute a global scale per-dimension
        total_observed = num_observed.sum(dim=0)
        denominator = torch.max(
            total_observed, torch.ones_like(total_observed)
        )
        default_scale = sum_observed.sum(dim=0) / denominator

        # then compute a per-item, per-dimension scale
        denominator = torch.max(num_observed, torch.ones_like(num_observed))
        scale = sum_observed / denominator

        # use per-batch scale when no element is observed
        # or when the sequence contains only zeros
        scale = torch.where(
            sum_observed > torch.zeros_like(sum_observed),
            scale,
            default_scale * torch.ones_like(num_observed),
        )

        return torch.max(scale, self.minimum_scale).detach()


class NOPScaler(Scaler):
    """
    The ``NOPScaler`` assigns a scale equals to 1 to each input item, i.e.,
    no scaling is applied upon calling the ``NOPScaler``.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compute_scale(
        self, data: torch.Tensor, observed_indicator: torch.Tensor
    ) -> torch.Tensor:
        if self.time_first:
            dim = 1
        else:
            dim = 2
        return torch.ones_like(data).mean(dim=dim)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import (
    constraints,
    NegativeBinomial,
    Poisson,
    Distribution,
)
from torch.distributions.utils import broadcast_all, lazy_property

from .utils import broadcast_shape


class ZeroInflatedDistribution(Distribution):
    """
    Generic Zero Inflated distribution.

    This can be used directly or can be used as a base class as e.g. for
    :class:`ZeroInflatedPoisson` and :class:`ZeroInflatedNegativeBinomial`.

    :param torch.Tensor gate: probability of extra zeros given via a Bernoulli distribution.
    :param TorchDistribution base_dist: the base distribution.
    """

    arg_constraints = {"gate": constraints.unit_interval}

    def __init__(self, gate, base_dist, validate_args=None):
        if base_dist.event_shape:
            raise ValueError(
                "ZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(
                    base_dist.event_shape
                )
            )
        batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
        self.gate = gate.expand(batch_shape)
        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()

        super().__init__(batch_shape, event_shape, validate_args)

    @property
    def support(self):
        return self.base_dist.support

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        gate, value = broadcast_all(self.gate, value)
        log_prob = (-gate).log1p() + self.base_dist.log_prob(value)
        log_prob = torch.where(
            value == 0, (gate + log_prob.exp()).log(), log_prob
        )
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
        return samples

    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean**2 + self.base_dist.variance
        ) - (self.mean) ** 2

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape)
        base_dist = self.base_dist.expand(batch_shape)
        ZeroInflatedDistribution.__init__(
            new, gate, base_dist, validate_args=False
        )
        new._validate_args = self._validate_args
        return new


class ZeroInflatedPoisson(ZeroInflatedDistribution):
    """
    A Zero Inflated Poisson distribution.

    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor rate: rate of poisson distribution.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "rate": constraints.positive,
    }
    support = constraints.nonnegative_integer

    def __init__(self, gate, rate, validate_args=None):
        base_dist = Poisson(rate=rate, validate_args=False)
        base_dist._validate_args = validate_args

        super().__init__(gate, base_dist, validate_args=validate_args)

    @property
    def rate(self):
        return self.base_dist.rate


class ZeroInflatedNegativeBinomial(ZeroInflatedDistribution):
    """
    A Zero Inflated Negative Binomial distribution.

    :param torch.Tensor gate: probability of extra zeros.
    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the half open interval [0, 1).
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "total_count": constraints.greater_than_eq(0),
        "probs": constraints.half_open_interval(0.0, 1.0),
        "logits": constraints.real,
    }
    support = constraints.nonnegative_integer

    def __init__(
        self, gate, total_count, probs=None, logits=None, validate_args=None
    ):
        base_dist = NegativeBinomial(
            total_count=total_count,
            probs=probs,
            logits=logits,
            validate_args=False,
        )
        base_dist._validate_args = validate_args

        super().__init__(gate, base_dist, validate_args=validate_args)

    @property
    def total_count(self):
        return self.base_dist.total_count

    @property
    def probs(self):
        return self.base_dist.probs

    @property
    def logits(self):
        return self.base_dist.logits



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
import time

import mxnet as mx
from mxnet import gluon
import numpy as np


class GMMModel(gluon.HybridBlock):
    """
    log p(x) >= E_{q(z|x)}[ log p(x|z) + log p(z) - log q(z|x) ]
    p(z) = unif
    p(x|z) = N(x; mu_z, (kR_z'kR_z)^-1)
    q(z|x) = softmax(-0.5*(kR_z(x-mu_z))^2 +logdet(kR_z) -0.5d*log(2*pi))
    mu_z = E_{q(z|x)} x
    cov_z = E_{q(z|x)} x^2 - (E_{q(z|x)} x)^2
    kR_z = inv(choL(cov_z))

    Shapes
    ===
    x: (batch_size, input_dim)
    log_marg: (batch_size,)
    qz: (batch_size, num_clusters)
    """

    def __init__(
        self,
        *data_template,
        num_clusters,
        log_prior_=None,
        mu_=None,
        kR_=None,
        lr_mult=None,
        ctx=None,
        hybridize=None,
    ):
        super().__init__()
        self.input_dim = data_template[0].shape[1]
        self.num_clusters = num_clusters

        with self.name_scope():
            self.log_prior_ = self.params.get(
                "log_prior_",
                shape=(num_clusters,),
                lr_mult=lr_mult,
                init=mx.init.Constant(np.log(1 / self.num_clusters))
                if log_prior_ is None
                else mx.init.Constant(log_prior_),
            )

            self.mu_ = self.params.get(
                "mu_",
                shape=(self.num_clusters, self.input_dim),
                lr_mult=lr_mult,
                init=None if mu_ is None else mx.init.Constant(mu_),
            )

            self.kR_ = self.params.get(
                "kR_",
                shape=(self.num_clusters, self.input_dim, self.input_dim),
                lr_mult=lr_mult,
                init=None if kR_ is None else mx.init.Constant(kR_),
            )

        if hybridize:
            self.hybridize()

        self.initialize(ctx=ctx)
        self(*[mx.nd.array(x, ctx=ctx) for x in data_template])

    @staticmethod
    def _get_dx_(F, x, mu_):
        """@Return (batch_size, num_clusters, input_dim)"""
        return F.broadcast_minus(x.expand_dims(1), mu_)

    @staticmethod
    def _get_Rx_(F, dx_, kR_):
        """@Return (batch_size, num_clusters, input_dim)"""
        kR_expand_0 = F.broadcast_like(
            kR_.expand_dims(0), dx_, lhs_axes=(0,), rhs_axes=(0,)
        )  # (batch_size, num_clusters, input_dim, input_dim)

        Rx_ = F.batch_dot(kR_expand_0, dx_.expand_dims(-1)).squeeze(axis=-1)
        return Rx_

    def hybrid_forward(self, F, x, log_prior_, mu_, kR_):
        """E-step computes log_marginal and q(z|x)"""
        dx_ = self._get_dx_(F, x, mu_)
        Rx_ = self._get_Rx_(F, dx_, kR_)

        log_conditional = (
            -0.5 * (Rx_**2).sum(axis=-1)
            - 0.5 * self.input_dim * np.log(2 * np.pi)
            + F.linalg.slogdet(kR_)[1]
        )  # (batch, num_clusters)

        log_complete = F.broadcast_add(
            log_conditional, log_prior_.log_softmax()
        )
        log_incomplete = F.log(F.exp(log_complete).sum(axis=1))

        qz = log_complete.softmax(axis=1)

        return log_incomplete, qz

    @staticmethod
    def m_step(x, qz):
        """M-step computes summary statistics in numpy"""
        x = x.astype("float64")
        qz = qz.astype("float64")

        nz = qz.sum(axis=0)  # (num_clusters,)
        sum_x = (qz[:, :, None] * x[:, None, :]).sum(axis=0)
        sum_x2 = (
            qz[:, :, None, None] * (x[:, None, :, None] @ x[:, None, None, :])
        ).sum(axis=0)
        return nz, sum_x, sum_x2


class GMMTrainer:
    """trainer based on M-step summary statistics
    can add mini-batch statistics for a full-batch update
    """

    def __init__(self, model, pseudo_count=0.1, jitter=1e-6):
        self.model = model
        self.pseudo_count = pseudo_count
        self.jitter = jitter
        self.zero_stats()

    def zero_stats(self):
        self.nz = self.pseudo_count * np.ones(self.model.num_clusters)
        self.sum_x = np.zeros(self.model.input_dim)
        self.sum_x2 = (
            np.eye(self.model.input_dim) * self.jitter * self.pseudo_count
        )

    def add(self, x):
        log_incomplete, qz = self.model(mx.nd.array(x))
        nz, sum_x, sum_x2 = self.model.m_step(x, qz.asnumpy())
        self.nz = self.nz + nz
        self.sum_x = self.sum_x + sum_x
        self.sum_x2 = self.sum_x2 + sum_x2
        return log_incomplete

    def update(self):
        mu_ = self.sum_x / self.nz[:, None]
        Ex2 = self.sum_x2 / self.nz[:, None, None]
        cov_ = Ex2 - mu_[:, :, None] @ mu_[:, None, :]
        kR_ = np.linalg.inv(np.linalg.cholesky(cov_))

        self.model.log_prior_.set_data(np.log(self.nz / self.nz.sum()))
        self.model.mu_.set_data(mu_)
        self.model.kR_.set_data(kR_)

        self.zero_stats()

    def __call__(self, x):
        self.add(x)
        self.update()


def infer_lambda(model, *_, xmin, xmax):
    """infer lambda and intercept based on linear fitting at the base points"""
    x = np.linspace(xmin, xmax).reshape((-1, 1))
    y = np.ravel(model(mx.nd.array(x))[0].asnumpy())
    slope, intercept = np.polyfit(np.ravel(x), np.ravel(y), 1)
    return -slope


def elapsed(collection):
    """similar to enumerate but prepend elapsed time since loop starts"""
    tic = time.time()
    for x in collection:
        yield time.time() - tic, x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Literal, TextIO
import yaml
from tsbench.config import TrainConfig


def iterate_configurations(
    configs: List[Dict[str, Any]], skip: int = 0
) -> Iterator[Dict[str, Any]]:
    """
    Provides an iterator over the configurations, ignoring the first `skip` configurations.

    Args:
        configs: The configurations to iterate over.
        skip: The number of configurations to skip at the beginning.

    Yields:
        The configuration.
    """
    for i, config_item in enumerate(configs):
        if i < skip:
            print(f">>> Skipping configuration {i+1}/{len(configs)}")
            continue
        print(f">>> Running configuration {i+1}/{len(configs)}")
        yield config_item


def generate_configurations(path: Path) -> List[Any]:
    """
    Generates the hyperparameter configuration for the configuration(s) found at the specified
    path.

    Args:
        path: Either a file or directory from to read the configuration(s). If a directory is
            given, all YAML files are read recursively.

    Returns:
        The configurations.
    """
    if path.is_file():
        with path.open() as f:
            return _generate_configurations(f)

    all_configurations = []
    for file in path.glob("**/*.yaml"):
        with file.open() as f:
            all_configurations.extend(_generate_configurations(f))
    return all_configurations


def explode_key_values(
    primary_key: str,
    mapping: Dict[str, List[Dict[Literal["key", "values"], Any]]],
    process_key: Callable[[str, str], str] = lambda _, s: s,
) -> List[Dict[str, Any]]:
    """
    Explodes a mapping from primary keys to a list of key to value mappings into independent
    configurations.
    """
    all_combinations = {
        primary: itertools.product(
            *[
                [(option["key"], value) for value in option["values"]]
                for option in choices
            ]
        )
        if choices
        else []
        for primary, choices in mapping.items()
    }

    # Generate configs
    configs = []
    for primary, combinations in all_combinations.items():
        if not combinations:
            configs.append({primary_key: primary})
        for item in combinations:
            primary_config = {primary_key: primary}
            for key, value in item:
                if isinstance(key, (list, tuple)):
                    primary_config.update(
                        {
                            process_key(primary, k): v
                            for k, v in zip(key, value)
                        }
                    )
                else:
                    primary_config[process_key(primary, key)] = value
            configs.append(primary_config)

    # Explode repetitions
    result = []
    for config in configs:
        if "__repeat__" in config:
            for _ in range(config["__repeat__"]):
                result.append(
                    {k: v for k, v in config.items() if k != "__repeat__"}
                )
        else:
            result.append(config)

    return result


def _generate_configurations(config: TextIO) -> List[Any]:
    # First, we parse the config file
    options = yaml.safe_load(config)
    seeds = options["seeds"]
    datasets = options["datasets"]

    def process_key(model: str, key: str) -> str:
        if key in TrainConfig.training_hyperparameters():
            return key
        return f"{model}_{key}"

    configs = explode_key_values("model", options["models"], process_key)

    # Explode all runs and save to file
    all_configurations = []
    for seed in seeds:
        for dataset in datasets:
            for model_config in configs:
                all_configurations.append(
                    {
                        "seed": seed,
                        "dataset": dataset,
                        **model_config,
                    }
                )

    return all_configurations



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import defaultdict
from typing import Dict, List, Optional
import numpy as np
import pygmo
from sklearn.preprocessing import QuantileTransformer, StandardScaler
from tsbench.config import Config, DatasetConfig, ModelConfig
from tsbench.evaluations.metrics import Performance
from ._base import Recommendation, Recommender
from ._factory import register_recommender
from .generator import CandidateGenerator


@register_recommender("greedy")
class GreedyRecommender(Recommender[ModelConfig]):
    """
    Recommender that selects a set of configurations differently in the one-dimensional and two-
    dimensional setting. In the one-dimensional setting, it greedily picks the configurations which
    provide the lowest joint error. In the multi-dimensional setting, it greedily picks the
    configurations which provide the lowest joint hypervolume error.
    """

    metrics: np.ndarray  # shape [num_datasets, num_models, num_objectives]
    model_indices: Dict[
        ModelConfig, int
    ]  # A map from model config to index in the metrics

    def __init__(
        self,
        objectives: List[str],
        focus: Optional[str] = None,
        generator: Optional[CandidateGenerator[ModelConfig]] = None,
        enforce_single_objective: bool = False,
    ):
        """
        Args:
            objectives: The list of performance metrics to minimize.
            focus: The metric to prefer. Must be either in the list of objectives. If not
                provided, the first metric to optimize is chosen.
            generator: The generator that generates configurations for recommendations. By default,
                this is the replay candidate generator.
            enforce_single_objective: Whether the greedy recommender should pick models greedily
                even in the multi-objective case and pick models alternately.
        """
        super().__init__(objectives, focus, generator)
        self.enforce_single_objective = enforce_single_objective

    def fit(
        self,
        configs: List[Config[ModelConfig]],
        performances: List[Performance],
    ) -> None:
        super().fit(configs, performances)

        # We need to sort by dataset to have the same ordering for each model config
        ordering = np.argsort([c.dataset.name() for c in configs])
        performance_df = Performance.to_dataframe(performances)

        # Extract all metrics
        metric_map = defaultdict(list)
        for i in ordering:
            metric_map[configs[i].model].append(
                performance_df.iloc[i][self.objectives].to_numpy(),  # type: ignore
            )

        # Build the properties
        self.metrics = np.stack(list(metric_map.values()), axis=1)
        self.model_indices = {model: i for i, model in enumerate(metric_map)}

        # If we are in the multi-objective setting, we have to apply dataset-level quantile
        # normalization of each objective. Otherwise, we perform standardization.
        if not self.enforce_single_objective and len(self.objectives) > 1:
            transformer = QuantileTransformer(
                n_quantiles=min(1000, self.metrics.shape[0])
            )
            self.metrics = np.stack(
                [
                    transformer.fit_transform(dataset_metrics)
                    for dataset_metrics in self.metrics
                ]
            )
        else:
            transformer = StandardScaler()
            self.metrics = np.stack(
                [
                    transformer.fit_transform(dataset_metrics)
                    for dataset_metrics in self.metrics
                ]
            )

    def recommend(
        self,
        dataset: DatasetConfig,
        candidates: Optional[List[ModelConfig]] = None,
        max_count: int = 10,
    ) -> List[Recommendation[ModelConfig]]:
        # Since recommendations are independent of the dataset, we could already compute the
        # result of this function in `fit`. However, we don't have access to `max_count` there.
        # First, we get the model configurations and check that they overlap with the ones we know.
        model_configs = self.generator.generate(candidates)
        assert all(
            c in self.model_indices for c in model_configs
        ), "Greedy recommender can only provide recommendations for known configurations."

        # Greedily pick configurations. For the multi-objective case, we need to compute the
        # hypervolume error. Hence, for each dataset, we need to compute the hypervolume of the
        # true Pareto front.
        if not self.enforce_single_objective and len(self.objectives) > 1:
            reference = np.ones(len(self.objectives))
            hypervolumes = np.array(
                [
                    pygmo.hypervolume(dataset_metrics).compute(reference)  # type: ignore
                    for dataset_metrics in self.metrics
                ]
            )

        available_choices = list(range(len(model_configs)))
        result = []
        while len(result) < max_count:
            # Pick the configuration which minimizes the joint error. For this, we need to compute
            # the error for all possible choices.
            errors = []
            for i, choice in enumerate(available_choices):
                # Compute the joint error with the model configuration of the choice
                all_configs = result + [model_configs[choice]]
                all_performances = self.metrics[
                    :, [self.model_indices[c] for c in all_configs]
                ]

                if self.enforce_single_objective or len(self.objectives) == 1:
                    # If we only consider a single objective, we can simply sum together the
                    # minimum observed values of the objective across the dataset
                    error = all_performances.min(1)[
                        :, i % len(self.objectives)
                    ].sum()
                else:
                    # Otherwise, we need to compute the hypervolumes for all datasets
                    reference = np.ones(len(self.objectives))
                    config_hypervolumes = np.array(
                        [
                            pygmo.hypervolume(  # type: ignore
                                dataset_performances,
                            ).compute(reference)
                            for dataset_performances in all_performances
                        ]
                    )
                    # And then compute the cumulative hypervolume error
                    error = (hypervolumes - config_hypervolumes).sum()  # type: ignore

                errors.append(error)

            # Get the index with the lowest loss
            lowest = np.argmin(errors)
            index = available_choices[lowest]
            del available_choices[lowest]
            result.append(model_configs[index])

        # Eventually return the configurations. The performance cannot be predicted by the greedy
        # surrogate.
        return [Recommendation(r, _dummy_performance()) for r in result]


def _dummy_performance() -> Performance:
    return Performance.from_dict(
        {
            mm: np.nan
            for m in Performance.metrics()
            for mm in [f"{m}_mean", f"{m}_std"]
        }
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Standard library imports
import functools
import itertools
import operator
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

# Third-party imports
import numpy as np
import pandas as pd

# First-party imports
from .common import DataEntry

ValueOrCallable = Union[Any, Callable]
Recipe = List[Tuple[str, Callable]]
Env = Dict[str, Any]


def resolve(val_or_callable: ValueOrCallable, context: Env, *args, **kwargs):
    if callable(val_or_callable):
        return val_or_callable(context, *args, **kwargs)
    elif isinstance(val_or_callable, str):
        return context[val_or_callable]
    else:
        return val_or_callable


def generate(
    length: int,
    recipe: Union[Callable, Recipe],
    start: pd.Timestamp,
    global_state: Optional[dict] = None,
    seed: int = 0,
    item_id_prefix: str = "",
) -> Iterator[DataEntry]:
    np.random.seed(seed)

    if global_state is None:
        global_state = {}

    if isinstance(recipe, list):
        for x in itertools.count():
            data: DataEntry = {}
            for k, f in recipe:
                data[k] = resolve(
                    f,
                    data,
                    length=length,
                    field_name=k,
                    global_state=global_state,
                )
            yield dict(**data, item_id=item_id_prefix + str(x), start=start)
    else:
        assert callable(recipe)
        for x in itertools.count():
            data = recipe(length=length, global_state=global_state)
            yield dict(**data, item_id=item_id_prefix + str(x), start=start)


def evaluate(
    funcs: Recipe, length: int, *args, global_state: dict = None, **kwargs
) -> Env:
    if global_state is None:
        global_state = {}

    if "length" in kwargs:
        del kwargs["length"]
    if "field_name" in kwargs:
        del kwargs["field_name"]
    if "global_state" in kwargs:
        del kwargs["global_state"]

    data: DataEntry = {}
    for k, f in funcs:
        try:
            data[k] = resolve(
                f,
                data,
                length=length,
                field_name=k,
                global_state=global_state,
                *args,
                **kwargs
            )
        except ValueError as e:
            raise ValueError('Error while evaluating key "{}"'.format(k), e)

    return data


def make_func(
    length: int, funcs: Recipe, global_state=None
) -> Callable[[int, Env], DataEntry]:
    if global_state is None:
        global_state = {}

    def f(length=length, global_state=global_state, *args, **kwargs):
        data = {}
        for k, f in funcs:
            data[k] = resolve(
                f,
                data,
                length=length,
                field_name=k,
                global_state=global_state,
                *args,
                **kwargs
            )
        return data

    return f


def take_as_list(iterator, num):
    return list(itertools.islice(iterator, num))


class Debug:
    def __init__(self, print_global=False) -> None:
        self.print_global = print_global

    def __call__(self, x: Env, global_state, **kwargs):
        print(x)
        if self.print_global:
            print(global_state)
        return 0


class Lifted:
    def __add__(self, other):
        return LiftedAdd(self, other)

    def __radd__(self, other):
        return LiftedAdd(other, self)

    def __sub__(self, other):
        return LiftedSub(self, other)

    def __rsub__(self, other):
        return LiftedSub(other, self)

    def __mul__(self, other):
        return LiftedMul(self, other, operator.mul)

    def __rmul__(self, other):
        return LiftedMul(other, self, operator.mul)

    def __truediv__(self, other):
        return LiftedTruediv(self, other, operator.truediv)

    def __rtruediv__(self, other):
        return LiftedTruediv(other, self, operator.truediv)

    def __call__(
        self,
        x: Env,
        length: int,
        field_name: str,
        global_state: Dict,
        *args,
        **kwargs
    ):
        pass


class LiftedBinaryOp(Lifted):
    def __init__(self, left, right, op) -> None:
        self.left = left
        self.right = right
        self.op = op

    def __call__(self, *args, **kwargs):
        left = resolve(self.left, *args, **kwargs)
        right = resolve(self.right, *args, **kwargs)
        return self.op(left, right)


class LiftedAdd(LiftedBinaryOp):
    def __init__(self, left, right) -> None:
        super().__init__(left, right, operator.add)


class LiftedSub(LiftedBinaryOp):
    def __init__(self, left, right) -> None:
        super().__init__(left, right, operator.sub)


class LiftedMul(LiftedBinaryOp):
    def __init__(self, left, right) -> None:
        super().__init__(left, right, operator.mul)


class LiftedTruediv(LiftedBinaryOp):
    def __init__(self, left, right) -> None:
        super().__init__(left, right, operator.truediv)


class RandomGaussian(Lifted):
    def __init__(
        self, stddev: ValueOrCallable = 1.0, shape: Sequence[int] = (0,)
    ) -> None:
        self.stddev = stddev
        self.shape = shape

    def __call__(self, x: Env, length: int, *args, **kwargs):
        stddev = resolve(self.stddev, x, length, *args, **kwargs)
        s = np.array(self.shape)
        s[s == 0] = length
        return stddev * np.random.randn(*s)


# Binary recipe that returns 1 if date is in holidays list and 0 otherwise
class BinaryHolidays(Lifted):
    # TODO: holidays is type List[datetime.date]
    def __init__(self, dates: List[pd.Timestamp], holidays: List[Any]) -> None:
        self.dates = dates
        self.holidays = holidays

    def __call__(self, *args, **kwargs):
        length = len(self.dates)
        out = np.ones(length)
        for i, date in enumerate(self.dates):
            # Convert to string to check if inside of holidays datatime.date
            if date.date() in self.holidays:
                out[i] = 1.0
            else:
                out[i] = 0.0
        return out


class RandomBinary(Lifted):
    def __init__(self, prob: ValueOrCallable = 0.1) -> None:
        self.prob = prob

    def __call__(self, x: Env, length: int, *args, **kwargs):
        prob = resolve(self.prob, x, length, *args, **kwargs)
        return 1.0 * (np.random.rand(length) < prob)


class RandomSymmetricDirichlet(Lifted):
    def __init__(
        self, alpha: ValueOrCallable = 1.0, shape: Sequence[int] = (0,)
    ) -> None:
        self.alpha = alpha
        self.shape = shape

    def __call__(self, x, length, *args, **kwargs):
        alpha = resolve(self.alpha, x, length, *args, **kwargs)
        s = np.array(self.shape)
        s[s == 0] = length
        return np.random.dirichlet(alpha * np.ones(s))


class BinaryMarkovChain(Lifted):
    def __init__(
        self, one_to_zero: ValueOrCallable, zero_to_one: ValueOrCallable
    ) -> None:
        self.one_to_zero = one_to_zero
        self.zero_to_one = zero_to_one

    def __call__(self, x: Env, length: int, *args, **kwargs):
        probs = np.zeros(2)
        probs[0] = resolve(self.zero_to_one, x, length, *args, **kwargs)
        probs[1] = resolve(self.one_to_zero, x, length, *args, **kwargs)
        out = np.ones(length, dtype=np.int)  # initial state is 1
        uu = np.random.rand(length)
        for i in range(1, length):
            if uu[i] < probs[out[i - 1]]:
                out[i] = 1 - out[i - 1]
            else:
                out[i] = out[i - 1]
        return out


class Constant(Lifted):
    def __init__(self, constant) -> None:
        self.constant = constant

    def __call__(self, *args, **kwargs):
        return self.constant


class ConstantVec(Lifted):
    def __init__(self, constant: ValueOrCallable) -> None:
        self.constant = constant

    def __call__(self, x: Env, length: int, *args, **kwargs):
        constant = resolve(self.constant, x, length, *args, **kwargs)
        return constant * np.ones(length)


class NormalizeMax(Lifted):
    def __init__(self, input) -> None:
        self.input = input

    def __call__(self, x: Env, *args, **kwargs):
        inp = resolve(self.input, x, *args, kwargs)
        return inp / np.max(inp)


class OnesLike(Lifted):
    def __init__(self, other) -> None:
        self.other = other

    def __call__(self, x, length, *args, **kwargs):
        other = resolve(self.other, x, length, **kwargs)
        return np.ones_like(other)


class LinearTrend(Lifted):
    def __init__(self, slope: ValueOrCallable = 1.0) -> None:
        self.slope = slope

    def __call__(self, x, length, *args, **kwargs):
        slope = resolve(self.slope, x, length, *args, **kwargs)
        return slope * np.arange(length) / length


class RandomCat:
    def __init__(
        self,
        cardinalities: List[int],
        prob_fun: Callable = RandomSymmetricDirichlet(alpha=1.0, shape=(0,)),
    ) -> None:
        self.cardinalities = cardinalities
        self.prob_fun = prob_fun

    def __call__(self, x, field_name, global_state, **kwargs):
        if field_name not in global_state:
            probs = [self.prob_fun(x, length=c) for c in self.cardinalities]
            global_state[field_name] = probs
        probs = global_state[field_name]
        cats = np.array(
            [
                np.random.choice(np.arange(len(probs[i])), p=probs[i])
                for i in range(len(probs))
            ]
        )
        return cats


class Lag(Lifted):
    def __init__(
        self,
        input: ValueOrCallable,
        lag: ValueOrCallable = 0,
        pad_const: int = 0,
    ) -> None:
        self.input = input
        self.lag = lag
        self.pad_const = pad_const

    def __call__(self, x, *args, **kwargs):
        feat = resolve(self.input, x, *args, **kwargs)
        lag = resolve(self.lag, x, *args, **kwargs)

        if lag > 0:
            lagged_feat = np.concatenate(
                (self.pad_const * np.ones(lag), feat[:-lag])
            )
        elif lag < 0:
            lagged_feat = np.concatenate(
                (feat[-lag:], self.pad_const * np.ones(-lag))
            )

        else:
            lagged_feat = feat
        return lagged_feat


class ForEachCat(Lifted):
    def __init__(self, fun, cat_field="cat", cat_idx=0) -> None:
        self.fun = fun
        self.cat_field = cat_field
        self.cat_idx = cat_idx

    def __call__(
        self,
        x: Env,
        length: int,
        field_name: str,
        global_state: Dict,
        *args,
        **kwargs
    ):
        c = x[self.cat_field][self.cat_idx]
        if field_name not in global_state:
            global_state[field_name] = np.empty(
                len(global_state[self.cat_field][self.cat_idx]),
                dtype=np.object,
            )
        if global_state[field_name][c] is None:
            global_state[field_name][c] = self.fun(
                x, length=length, field_name=field_name, *args, **kwargs
            )
        return global_state[field_name][c]


class Eval(Lifted):
    def __init__(self, expr: str) -> None:
        self.expr = expr

    def __call__(self, x: Env, length: int, *args, **kwargs):
        return eval(self.expr, globals(), dict(x=x, length=length, **kwargs))


class SmoothSeasonality(Lifted):
    def __init__(
        self, period: ValueOrCallable, phase: ValueOrCallable
    ) -> None:
        self.period = period
        self.phase = phase

    def __call__(self, x: Env, length: int, *args, **kwargs):
        period = resolve(self.period, x, length, *args, **kwargs)
        phase = resolve(self.phase, x, length, *args, **kwargs)
        return (
            np.sin(2.0 / period * np.pi * (np.arange(length) + phase)) + 1
        ) / 2.0


class Add(Lifted):
    def __init__(self, inputs: List[ValueOrCallable]) -> None:
        self.inputs = inputs

    def __call__(self, x: Env, length: int, *args, **kwargs):
        return sum(
            [resolve(k, x, length, *args, **kwargs) for k in self.inputs]
        )


class Mul(Lifted):
    def __init__(self, inputs) -> None:
        self.inputs = inputs

    def __call__(self, x: Env, length: int, *args, **kwargs):
        return functools.reduce(
            operator.mul,
            [resolve(k, x, length, *args, **kwargs) for k in self.inputs],
        )


class NanWhere(Lifted):
    def __init__(
        self, source: ValueOrCallable, nan_indicator: ValueOrCallable
    ) -> None:
        self.source = source
        self.nan_indicator = nan_indicator

    def __call__(self, x: Env, length: int, *args, **kwargs):
        source = resolve(self.source, x, length, *args, **kwargs)
        nan_indicator = resolve(self.nan_indicator, x, length, *args, **kwargs)
        out = source.copy()
        out[nan_indicator == 1] = np.nan
        return out


class OneMinus(Lifted):
    def __init__(self, source: ValueOrCallable) -> None:
        self.source = source

    def __call__(self, x: Env, length: int, *args, **kwargs):
        value = resolve(self.source, x, length, *args, **kwargs)
        return 1 - value


class Concatenate(Lifted):
    def __init__(self, inputs: List[ValueOrCallable], axis: int = 0) -> None:
        self.inputs = inputs
        self.axis = axis

    def __call__(self, x: Env, length: int, *args, **kwargs):
        inputs = [resolve(z, x, length, **kwargs) for z in self.inputs]
        return np.concatenate(inputs, self.axis)


class Stack(Lifted):
    def __init__(self, inputs: List[ValueOrCallable]) -> None:
        self.inputs = inputs

    def __call__(self, x: Env, length: int, *args, **kwargs):
        inputs = [resolve(z, x, length, **kwargs) for z in self.inputs]
        return np.stack(inputs, axis=0)


class StackPrefix(Lifted):
    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    def __call__(self, x: Env, length: int, *args, **kwargs):
        inputs = [v for k, v in x.items() if k.startswith(self.prefix)]
        return np.stack(inputs, axis=0)


class Ref(Lifted):
    def __init__(self, field_name: str) -> None:
        self.field_name = field_name

    def __call__(self, x: Env, length: int, *args, **kwargs):
        return x[self.field_name]


class RandomUniform(Lifted):
    def __init__(
        self,
        low: ValueOrCallable = 0.0,
        high: ValueOrCallable = 1.0,
        shape=(0,),
    ) -> None:
        self.low = low
        self.high = high
        self.shape = shape

    def __call__(self, x: Env, length: int, *args, **kwargs):
        low = resolve(self.low, x, length, *args, **kwargs)
        high = resolve(self.high, x, length, *args, **kwargs)
        s = np.array(self.shape)
        s[s == 0] = length
        return np.random.uniform(low, high, s)


class RandomInteger(Lifted):
    def __init__(
        self,
        low: ValueOrCallable,
        high: ValueOrCallable,
        shape: Optional[Sequence[int]] = (0,),
    ) -> None:
        self.low = low
        self.high = high
        self.shape = shape

    def __call__(self, x: Env, length: int, *args, **kwargs):
        low = resolve(self.low, x, length, *args, **kwargs)
        high = resolve(self.high, x, length, *args, **kwargs)
        if self.shape is not None:
            s = np.array(self.shape)
            s[s == 0] = length
            return np.random.randint(low, high, s)
        else:
            return np.random.randint(low, high)


class RandomChangepoints(Lifted):
    def __init__(self, max_num_changepoints: ValueOrCallable) -> None:
        self.max_num_changepoints = max_num_changepoints

    def __call__(self, x: Env, length: int, *args, **kwargs):
        max_num_changepoints = resolve(
            self.max_num_changepoints, x, length, *args, **kwargs
        )
        num_changepoints = np.random.randint(0, max_num_changepoints + 1)
        change_idx = np.sort(
            np.random.randint(low=1, high=length - 1, size=(num_changepoints,))
        )
        change_ranges = np.concatenate([change_idx, [length]])
        out = np.zeros(length, dtype=np.int)
        for i in range(0, num_changepoints):
            out[change_ranges[i] : change_ranges[i + 1]] = i + 1
        return out


class Repeated(Lifted):
    def __init__(self, pattern: ValueOrCallable) -> None:
        self.pattern = pattern

    def __call__(self, x: Env, length: int, *args, **kwargs):
        pattern = resolve(self.pattern, x, length, **kwargs)
        repeats = length // len(pattern) + 1
        out = np.tile(pattern, (repeats,))
        return out[:length]


class Convolve(Lifted):
    def __init__(
        self, input: ValueOrCallable, filter: ValueOrCallable
    ) -> None:
        self.filter = filter
        self.input = input

    def __call__(self, x: Env, length: int, *args, **kwargs):
        fil = resolve(self.filter, x, length, **kwargs)
        inp = resolve(self.input, x, length, **kwargs)
        out = np.convolve(inp, fil, mode="same")
        return out


class Dilated(Lifted):
    def __init__(self, source: Callable, dilation: int) -> None:
        self.source = source
        self.dilation = dilation

    def __call__(self, x: Env, length: int, *args, **kwargs):
        inner = self.source(x, length // self.dilation + 1, **kwargs)
        out = np.repeat(inner, self.dilation)
        return out[:length]


class Choose(Lifted):
    def __init__(
        self, options: ValueOrCallable, selector: ValueOrCallable
    ) -> None:
        self.options = options
        self.selector = selector

    def __call__(self, x, length, **kwargs):
        options = resolve(self.options, x, length, **kwargs)
        selector = resolve(self.selector, x, length, **kwargs)
        e = np.eye(options.shape[0])
        out = np.sum(e[selector] * options.T, axis=1)
        return out


class EvalRecipe(Lifted):
    def __init__(self, recipe: Recipe, op: ValueOrCallable) -> None:
        self.recipe = recipe
        self.op = op

    def __call__(self, x: Env, *args, **kwargs):
        xx = evaluate(self.recipe, *args, **kwargs)
        return resolve(self.op, xx, *args, **kwargs)
